import time
import unittest

import numpy as np

from ._term_id import SimpleTermId, DefaultTermId


@unittest.skip('The bench is not run routinely.')
class TestTermIdBench(unittest.TestCase):
    """
    A naive benchmark to compare time required to create a set from different TermId implementations.
    """

    def test_bench_create_set(self):
        n_iter = 10
        n_elements = 10_000
        n_concepts = 1_000_000

        corpus = np.array(['HP:' + str(i).rjust(7, '0') for i in range(n_concepts)])
        simple = np.fromiter(map(lambda curie: SimpleTermId(curie, 2), corpus), dtype=object)
        default = np.fromiter(map(lambda curie: DefaultTermId(curie, 2), corpus), dtype=object)

        simple_elapsed = []
        default_elapsed = []
        for _ in range(n_iter):
            idx = np.random.choice(n_concepts, size=n_elements, replace=True)
            simple_sel = simple[idx]
            elapsed = time_it(set, simple_sel)
            simple_elapsed.append(elapsed)

            default_sel = default[idx]
            elapsed = time_it(set, default_sel)
            default_elapsed.append(elapsed)

        s = np.array(simple_elapsed) / 1e6  # To convert to ms
        d = np.array(default_elapsed) / 1e6  # To convert to ms

        print(f'Simple: {s.mean():.3f} ±{s.std():.3f}ms')
        print(f'Default: {d.mean():.3f} ±{d.std():.3f}ms')


def time_it(f, *args):
    start = time.time_ns()
    f(*args)
    return time.time_ns() - start



import typing
from warnings import warn

from hpotk.model import TermId, CURIE_OR_TERM_ID
from hpotk.graph import OntologyGraph, GraphAware


def get_ancestors(g: typing.Union[GraphAware, OntologyGraph],
                  source: CURIE_OR_TERM_ID,
                  include_source: bool = False) -> typing.FrozenSet[TermId]:
    """
    Get all ancestor :class:`TermId`\\ (s). of the `source` term (parents, grandparents, great-grandparents etc.)..

    The method raises a :class:`ValueError` if inputs do not meet the requirement described below.

    :param g: the ontology graph or a graph-aware entity
    :param source: `:class:`TermId` or a term ID curie as a `str (e.g. `HP:1234567`)
    :param include_source: whether to include the `source` term in the resulting set
    :return: a `frozenset` with ancestor :class:`TermId`\\ (s).
    """
    # TODO[v1.0.0] - remove the deprecated method
    warn('The method is deprecated and will be removed in v1.0.0. Use `get_ancestors` of the graph instead',
         DeprecationWarning, stacklevel=2)
    # Check
    g = _check_ontology_graph_is_available(g)
    source = _check_curie_or_term_id(source)

    # Init
    builder: set[TermId] = set()
    if include_source:
        builder.add(source)

    builder.update(g.get_ancestors(source))

    return frozenset(builder)


def get_parents(g: typing.Union[GraphAware, OntologyGraph],
                source: CURIE_OR_TERM_ID,
                include_source: bool = False) -> typing.FrozenSet[TermId]:
    """
    Get :class:`TermId`\\ (s). of the direct parents of the `source` term.

    The method raises a :class:`ValueError` if inputs do not meet the requirement described below.

    :param g: the ontology graph or a graph-aware entity
    :param source: :class:`TermId` or a term ID curie as a :class:`str (e.g. `HP:1234567`)
    :param include_source:  whether to include the `source` term ID(s) in the results
    :return: a :class:`frozenset` with parent `TermId`s
    """
    # TODO[v1.0.0] - remove the deprecated method
    warn('The method is deprecated and will be removed in v1.0.0. Use `get_parents` of the graph instead',
         DeprecationWarning, stacklevel=2)
    # Check
    g = _check_ontology_graph_is_available(g)
    source = _check_curie_or_term_id(source)

    # Init
    builder: typing.Set[TermId] = set()
    if include_source:
        builder.add(source)

    builder.update(g.get_parents(source))

    return frozenset(builder)


def get_descendants(g: typing.Union[GraphAware, OntologyGraph],
                    source: CURIE_OR_TERM_ID,
                    include_source: bool = False) -> typing.FrozenSet[TermId]:
    """
    Get all descendants `TermId`s of the `source` term (children, grandchildren, great-grandchildren etc.)..

    The method raises a `ValueError` if inputs do not meet the requirement described below.

    :param g: the ontology graph or a graph-aware entity
    :param source: `TermId` or a term ID curie as a `str (e.g. `HP:1234567`)
    :param include_source:  whether to include the `source` term ID(s) in the results
    :return: a :class:`frozenset` with descendants `TermId`s
    """
    # TODO[v1.0.0] - remove the deprecated method
    warn('The method is deprecated and will be removed in v1.0.0. Use `get_descendants` of the graph instead',
         DeprecationWarning, stacklevel=2)
    # Check
    g = _check_ontology_graph_is_available(g)
    source = _check_curie_or_term_id(source)

    # Init
    builder: set[TermId] = set()
    if include_source:
        builder.add(source)

    builder.update(g.get_descendants(source))

    return frozenset(builder)


def get_children(g: typing.Union[GraphAware, OntologyGraph],
                 source: CURIE_OR_TERM_ID,
                 include_source: bool = False) -> typing.FrozenSet[TermId]:
    """
    Get `TermId`s of the direct children of the `source` term.

    The method raises a `ValueError` if inputs do not meet the requirement described below.

    :param g: the ontology graph or a graph-aware entity
    :param source: `TermId` or a CURIE `str` (e.g. `HP:1234567`)
    :param include_source: whether to include the `source` term in the results
    :return: an iterable with child `TermId`s
    """
    # TODO[v1.0.0] - remove the deprecated method
    warn('The method is deprecated and will be removed in v1.0.0. Use `get_children` of the graph instead',
         DeprecationWarning, stacklevel=2)
    # Check
    g = _check_ontology_graph_is_available(g)
    source = _check_curie_or_term_id(source)

    builder: set[TermId] = set()
    if include_source:
        builder.add(source)

    builder.update(g.get_children(source))

    return frozenset(builder)


def exists_path(g: typing.Union[GraphAware, OntologyGraph],
                source: CURIE_OR_TERM_ID,
                destination: CURIE_OR_TERM_ID) -> bool:
    """
    Return `True` if `destination` is an ancestor of the `source` term.

    The path does *not* exists if `source` and `destination` are the same term.

    :param g: the ontology graph or a graph-aware entity
    :param source: `TermId` or a CURIE `str` (e.g. `HP:1234567`)
    :param destination: `TermId` or a CURIE `str` (e.g. `HP:1234567`)
    :return: `True` if a path exists from `source` to `destination`
    """
    g = _check_ontology_graph_is_available(g)
    source = _check_curie_or_term_id(source)
    destination = _check_curie_or_term_id(destination)

    if source == destination:
        return False

    for ancestor in get_ancestors(g, source):
        if ancestor == destination:
            return True

    return False


def _check_ontology_graph_is_available(g):
    if isinstance(g, OntologyGraph):
        pass
    elif isinstance(g, GraphAware):
        g = g.graph
    else:
        raise ValueError(f'`g` must implement `OntologyGraph` or `GraphAware` but got {type(g)}')
    return g


def _check_curie_or_term_id(source: CURIE_OR_TERM_ID) -> TermId:
    if isinstance(source, str):
        source = TermId.from_curie(source)
    elif isinstance(source, TermId):
        pass
    else:
        raise ValueError(f'`source` must be `TermId` or a CURIE `str` but got {type(source)}')
    return source



import abc
import typing
import warnings

import numpy as np
from collections import deque


class ShapedMixin(metaclass=abc.ABCMeta):

    @property
    @abc.abstractmethod
    def shape(self):
        pass


class CsrMatrixBuilder(ShapedMixin):
    """
    The builder has been deprecated due to poor performance and will be removed in v1.0.0.
    """

    def __init__(self, shape: typing.Tuple[int, int]):
        _check_shape(shape)
        self._shape = shape
        self._row = np.zeros(shape=(shape[0] + 1,), dtype=int)
        self._col = deque([])
        self._data = deque([])
        warnings.warn('CsrMatrixBuilder has been deprecated and will be removed in v1.0.0',
                      DeprecationWarning, stacklevel=2)

    def __setitem__(self, key, value):
        if isinstance(key, tuple):
            if len(key) == 2:
                qrow, qcol = key
                _check_bounds(qrow, qcol, self._shape)

                start, end = self._row[qrow: qrow + 2]

                col_idx_adjustment, value_will_be_updated = 0, False  # Loop variables
                for i, col in enumerate(self._col):
                    # We iterate over all entries of the _indices deque since we cannot slice the deque
                    if start > i:
                        # Let's discard the first [0..row-1] rows.
                        continue
                    if i >= end:
                        # No need to iterate over the rows beyond the target row.
                        break

                    # Here we are at the relevant part of the `self._indices`.
                    # We determine if we need to update an existing value,
                    # or create a new entry at an approperiate place
                    if col < qcol:
                        col_idx_adjustment += 1
                        continue
                    elif col == qcol:
                        value_will_be_updated = True

                    break

                idx = start + col_idx_adjustment

                if value_will_be_updated:
                    self._data[idx] = value
                else:
                    # We are adding a new entry, so we must increment relevant _indptr entries
                    mask = np.arange(self._row.shape[0]) > qrow
                    self._row[mask] += 1
                    self._col.insert(idx, qcol)
                    self._data.insert(idx, value)

            else:
                raise ValueError(f'Setting {len(key)} dimensions but only 2D indexing is supported')
        else:
            raise IndexError(f'Unknown index type {type(key)}')

    @property
    def shape(self):
        return self._shape

    @property
    def row(self):
        return self._row

    @property
    def col(self):
        return self._col

    @property
    def data(self):
        return self._data

    def __repr__(self):
        return str(self)

    def __str__(self):
        return f"CsrMatrixBuilder(shape={self._shape}, indptr={self._row}, indices={self._col}, data={self._data})"


class ImmutableCsrMatrix(ShapedMixin):

    def __init__(self, row: typing.Sequence,
                 col: typing.Sequence,
                 data: typing.Sequence,
                 shape: typing.Tuple[int, int],
                 dtype=float):
        # Sanity checks
        _check_sequence_of_nonnegative_ints('row', row)
        _check_sequence_of_nonnegative_ints('col', col)
        if not isinstance(data, typing.Sequence):
            raise ValueError(f'data must be a sequence but was {type(type(data))}')

        _check_shape(shape)

        if len(row) - 1 != shape[0]:
            raise ValueError(f'row len {len(row) - 1} must be equal to number of rows {shape[0]}')

        if not isinstance(dtype, type):
            raise ValueError(f'dtype must be a type but was {type(dtype)}')

        # Store the state in numpy arrays
        self._row = np.array(row)
        self._col = np.array(col)
        self._data = np.array(data)
        self._shape = shape
        self._dtype = dtype
        self._default = self._default_for_dtype(dtype)

    def __getitem__(self, item):
        if isinstance(item, int):
            if 0 <= item < self._shape[0]:
                start_row, end_row = self._row[item: item + 2]
                row = np.full(shape=(self._shape[1],),
                              fill_value=self._default,
                              dtype=self._dtype)
                if start_row != end_row:
                    idxs = self._col[start_row: end_row]
                    vals = self._data[start_row: end_row]
                    row[idxs] = vals
                return row
            else:
                if 0 > item:
                    raise ValueError(f'Requested row #{item} but negative indexing is not supported')
                else:
                    raise IndexError(f'Row index {item} out of bounds for a {self._shape} matrix')
        elif isinstance(item, tuple):
            if len(item) == 2:
                qrow, qcol = item
                _check_bounds(qrow, qcol, self._shape)
                # +2 is safe since we check bounds above and self._indptr has n+1 elements for n x m matrix.
                start_row, end_row = self._row[qrow: qrow + 2]
                for i, col_idx in enumerate(self._col[start_row: end_row]):
                    if col_idx == qcol:
                        return self._data[start_row + i]
                return self._default
            else:
                raise ValueError(f'Requesting {len(item)} dimensions but only 2D indexing is supported')
        else:
            raise IndexError(f'Unknown index type {type(item)}')

    def col_indices_of_val(self, row: int, query):
        """
        Return indices of columns with matching query value in a given row.

        Raises IndexError if `row` is out of bounds.
        """
        if not (isinstance(row, int) and 0 <= row < self._shape[0]):
            raise IndexError(f'row must be an int in range [0, {self.shape[0]}) but was {row}')

        start_row, end_row = self._row[row: row + 2]
        value_idxs = self._col[start_row: end_row]
        if query == self._default:
            # We do not store indices of the default value in the matrix
            col_indices = np.arange(self._shape[1])
            row_mask = np.full(fill_value=True, shape=(self._shape[1],), dtype=bool)
            if len(value_idxs) != 0:
                row_mask[value_idxs] = False
                return col_indices[row_mask]
            else:
                return col_indices
        else:
            # We're getting indices of columns where data is equal to the query
            row_mask = self._data[start_row: end_row] == query
            return value_idxs[row_mask]

    @property
    def shape(self):
        return self._shape

    @staticmethod
    def _default_for_dtype(dtype):
        if dtype == float:
            return 0.
        elif dtype == int:
            return 0
        elif dtype == bool:
            return False


def _check_shape(shape):
    if not (isinstance(shape, tuple) and len(shape) == 2):
        raise ValueError(f'shape must be a tuple with two non-negative ints')
    _check_sequence_of_nonnegative_ints('shape', shape)


def _check_bounds(row, col, shape):
    # Check bounds
    if not (0 <= row < shape[0]):
        raise IndexError(f'Row index {row} out of bounds for a {shape} matrix')
    if not (0 <= col < shape[1]):
        raise IndexError(f'Column index {col} out of bounds for a {shape} matrix')


def _check_sequence_of_nonnegative_ints(name, vals):
    if not (isinstance(vals, (typing.Sequence, np.ndarray)) and all([val >= 0 for val in vals])):
        raise ValueError(f'{name} must be a sequence of ints')



import pytest

from hpotk import TermId, OntologyGraph
from hpotk.graph import IncrementalCsrGraphFactory
from ._hierarchical import HierarchicalIcTermIdSorting, HierarchicalEdgeTermIdSorting, EdgeSimilarityMeasure


def make_nodes_and_edges():
    """Prepare nodes and edges for testing."""
    # Nodes are sorted.
    nodes = [
        TermId.from_curie('HP:01'),
        TermId.from_curie('HP:010'),
        TermId.from_curie('HP:011'),
        TermId.from_curie('HP:0110'),
        TermId.from_curie('HP:02'),
        TermId.from_curie('HP:020'),
        TermId.from_curie('HP:021'),
        TermId.from_curie('HP:022'),
        TermId.from_curie('HP:03'),
        TermId.from_curie('HP:1'),
    ]

    # Edges for a graph with a single root node: HP:1
    edges = [
        # source -> destination
        (TermId.from_curie('HP:01'), TermId.from_curie('HP:1')),
        (TermId.from_curie('HP:010'), TermId.from_curie('HP:01')),
        (TermId.from_curie('HP:011'), TermId.from_curie('HP:01')),
        # This one is multi-parent.
        (TermId.from_curie('HP:0110'), TermId.from_curie('HP:010')),
        (TermId.from_curie('HP:0110'), TermId.from_curie('HP:011')),

        (TermId.from_curie('HP:02'), TermId.from_curie('HP:1')),
        (TermId.from_curie('HP:020'), TermId.from_curie('HP:02')),
        (TermId.from_curie('HP:021'), TermId.from_curie('HP:02')),
        (TermId.from_curie('HP:022'), TermId.from_curie('HP:02')),

        (TermId.from_curie('HP:03'), TermId.from_curie('HP:1')),
    ]
    return nodes, edges


@pytest.fixture
def ontology_graph() -> OntologyGraph:
    factory = IncrementalCsrGraphFactory()
    _, edges = make_nodes_and_edges()
    return factory.create_graph(edges)


class TestHierarchicalSimilaritySorting:

    @pytest.fixture
    def ic_sim_sorting(self, ontology_graph: OntologyGraph) -> HierarchicalIcTermIdSorting:
        ics = {
            'HP:1': 0.,

            'HP:01': 1.,
            'HP:010': 2., 'HP:011': 2.3, 'HP:0110': 3.2,

            'HP:02': 1.5,
            'HP:020': 2.4, 'HP:021': 2.5, 'HP:022': 2.8,

            'HP:03': 2.
        }

        def ic_source(term_id: TermId) -> float: return ics.get(term_id.value, 0.)

        return HierarchicalIcTermIdSorting(ontology_graph, ic_source)

    @pytest.mark.parametrize('curies, expected',
                             [
                                 (('HP:021', 'HP:03', 'HP:01', 'HP:02', 'HP:1', 'HP:022', 'HP:0110', 'HP:020'),
                                  ('HP:0110', 'HP:01', 'HP:020', 'HP:022', 'HP:02', 'HP:021', 'HP:1', 'HP:03')),

                                 # Does not preserve order at the moment :'(.
                                 (('HP:0110', 'HP:01', 'HP:020', 'HP:022', 'HP:02', 'HP:021', 'HP:1', 'HP:03'),
                                  ('HP:01', 'HP:0110', 'HP:021', 'HP:02', 'HP:022', 'HP:020', 'HP:03', 'HP:1')),

                                 (('HP:021',),
                                  ('HP:021',))
                             ]
                             )
    def test_argsort(self, curies, expected, ic_sim_sorting: HierarchicalIcTermIdSorting):
        term_ids = tuple(map(TermId.from_curie, curies))

        term_indices = ic_sim_sorting.argsort(term_ids)

        ordered = tuple(term_ids[idx].value for idx in term_indices)
        assert expected == ordered


class TestHierarchicalEdgeTermIdSorting:

    @pytest.fixture
    def edge_sim_sorting(self, ontology_graph: OntologyGraph) -> HierarchicalEdgeTermIdSorting:
        return HierarchicalEdgeTermIdSorting(ontology_graph)

    @pytest.mark.parametrize('curies, expected',
                             [
                                 (('HP:021', 'HP:03', 'HP:01', 'HP:02', 'HP:1', 'HP:022', 'HP:0110', 'HP:020'),
                                  ('HP:02', 'HP:021', 'HP:022', 'HP:020', 'HP:1', 'HP:03', 'HP:01', 'HP:0110')),

                                 # Does not preserve order at the moment :'(.
                                 (('HP:0110', 'HP:01', 'HP:020', 'HP:022', 'HP:02', 'HP:021', 'HP:1', 'HP:03'),
                                  ('HP:1', 'HP:01', 'HP:03', 'HP:02', 'HP:020', 'HP:022', 'HP:021', 'HP:0110')),

                                 (('HP:021',),
                                  ('HP:021',))
                             ]
                             )
    def test_argsort(self, curies, expected, edge_sim_sorting: HierarchicalEdgeTermIdSorting):
        term_ids = tuple(map(TermId.from_curie, curies))

        term_indices = edge_sim_sorting.argsort(term_ids)

        ordered = tuple(term_ids[idx].value for idx in term_indices)
        assert ordered == expected


class TestEdgeSimilarityMeasure:

    @pytest.fixture
    def edge_sim(self, ontology_graph: OntologyGraph) -> EdgeSimilarityMeasure:
        return EdgeSimilarityMeasure(ontology_graph)

    @pytest.mark.parametrize('left, right, common_ancestor, expected',
                             [
                                 ('HP:1', 'HP:1', 'HP:1', 0),
                                 ('HP:01', 'HP:01', 'HP:01', 0),
                                 ('HP:010', 'HP:0110', 'HP:010', 1),
                                 ('HP:011', 'HP:0110', 'HP:011', 1),
                                 ('HP:01', 'HP:0110', 'HP:01', 2),
                                 ('HP:1', 'HP:0110', 'HP:1', 3),
                                 ('HP:03', 'HP:01', 'HP:1', 2),
                                 ('HP:03', 'HP:02', 'HP:1', 2),
                                 ('HP:0110', 'HP:022', 'HP:1', 5),
                             ])
    def test_compute_edge_distance(self, left: str, right: str, common_ancestor: str, expected: int,
                                   edge_sim: EdgeSimilarityMeasure):
        left = TermId.from_curie(left)
        right = TermId.from_curie(right)
        common_ancestor = TermId.from_curie(common_ancestor)

        lr_dist, lr_term_id = edge_sim.calculate_edge_distance(left, right)
        assert lr_dist == expected
        rl_dist, rl_term_id = edge_sim.calculate_edge_distance(right, left)
        assert rl_dist == expected
        assert lr_term_id == rl_term_id
        assert lr_term_id == common_ancestor



import gzip
import io
import logging
import sys
import typing
import warnings
from urllib.request import urlopen


def looks_like_url(file: str) -> bool:
    """
   Checks if the `file` looks like a URL.

   :param file: file to check.
   :return: `True` if the `file` starts with `http://` or `https://`.
   """
    return file.startswith('http://') or file.startswith('https://')


def looks_gzipped(file: str) -> bool:
    """
    Checks file suffix to determine if it looks gzipped.

    :param file: file path to check.
    :return: `True` if the `file` ends with `.gz`.
    """
    return file.endswith('.gz')


def _parse_encoding(encoding, logger) -> str:
    if encoding is None:
        encoding = sys.getdefaultencoding()
        logger.debug(f'Using default encoding \'{encoding}\'')
    else:
        logger.debug(f'Using provided encoding \'{encoding}\'')
    return encoding


def open_text_io_handle_for_reading(fh: typing.Union[typing.IO, str],
                                    timeout: int = 30,
                                    encoding: str = None) -> typing.TextIO:
    """
    Open a `io.TextIO` file handle based on `fh`.

    :param fh: a `str` or `typing.IO` to read from. If `str`, then it should be a path to a local file or a URL
      of a remote resource. Either `http` or `https` protocols are supported. The content will be uncompressed
      on the fly if the file name ends with `.gz`. If `fh` is an IO wrapper, the function ensures we get a text wrapper
      that uses given encoding.
    :param timeout: timeout in seconds used when accessing a remote resource.
    :param encoding: encoding used to decode the input or the system preferred encoding if unset.
    :return: the :class:`io.TextIO` wrapper.
    """
    logger = logging.getLogger('hpotk.util')
    encoding = _parse_encoding(encoding, logger)

    logger.debug(f'Opening {fh}')
    if isinstance(fh, str):
        # Can be a path to local file or URL
        if looks_like_url(fh):
            logger.debug(f'Looks like a URL: {fh}')
            if not isinstance(timeout, int) or timeout <= 0:
                raise ValueError(f'If {fh} looks like URL then timeout {timeout} must be a positive `int`')
            logger.debug(f'Downloading with timeout={timeout}s')
            handle = urlopen(fh, timeout=timeout)
        else:
            logger.debug(f'Looks like a local file: {fh}')
            handle = open(fh, 'rb')

        if looks_gzipped(fh):
            logger.debug(f'Looks like a gzipped data, decompressing on the fly')
            return gzip.open(handle, mode='rt', newline='', encoding=encoding)
        else:
            logger.debug(f'Looks like decompressed data')
            return io.TextIOWrapper(handle, encoding=encoding)
    elif isinstance(fh, typing.IO):
        if isinstance(fh, typing.BinaryIO):
            logger.debug(f'Looks like a binary IO')
            return io.TextIOWrapper(fh, encoding=encoding)
        elif isinstance(fh, typing.TextIO):
            return fh
        else:
            raise ValueError(f'Unexpected type {type(fh)}')
    else:
        raise ValueError(f'Unexpected type {type(fh)}')


def open_text_io_handle(fh: typing.Union[typing.IO, str],
                        timeout: int = 30,
                        encoding: str = None) -> typing.TextIO:
    """
    Open a `io.TextIO` file handle based on `fh`.

    :param fh: a `str` or `typing.IO` to read from. If `str`, then it should be a path to a local file or a URL
      of a remote resource. Either `http` or `https` protocols are supported. The content will be uncompressed
      on the fly if the file name ends with `.gz`. If `fh` is an IO wrapper, the function ensures we get
      a text wrapper that uses given encoding.
    :param timeout: timeout in seconds used when accessing a remote resource.
    :param encoding: encoding used to decode the input or the system preferred encoding if unset.
    :return: the :class:`io.TextIO` wrapper.
    """
    # REMOVE(v1.0.0)
    warnings.warn('The method has been deprecated and will be removed in v1.0.0. '
                  'Use `open_text_io_handle_for_reading` instead', DeprecationWarning, stacklevel=2)
    return open_text_io_handle_for_reading(fh, timeout, encoding)


def open_text_io_handle_for_writing(fh: typing.Union[str, typing.IO],
                                    encoding: str = None) -> typing.TextIO:
    """
    Open a `io.TextIO` file handle based on `fpath`.

    :param fh: a `str` with a path to a local file The content will be compressed on the fly if the file name ends
      with `.gz`.
    :param encoding: encoding used to encode the output or the system preferred encoding if unset.
    :return: a :class:`io.TextIO` wrapper.
    """
    logger = logging.getLogger('hpotk.util')
    encoding = _parse_encoding(encoding, logger)

    if isinstance(fh, str):
        if looks_gzipped(fh):
            logger.debug(f'Looks like a gzipped data, compressing on the fly')
            return gzip.open(fh, mode='wt', newline='', encoding=encoding)
        else:
            return open(fh, 'w')
    elif isinstance(fh, typing.BinaryIO):
        logger.debug(f'Looks like a binary IO')
        return io.TextIOWrapper(fh, encoding=encoding)
    elif isinstance(fh, typing.TextIO):
        return fh
    else:
        raise ValueError(f'Unexpected type {type(fh)}')



import typing

import pytest

import hpotk
from ._csr_idx_graph import CsrIndexedOntologyGraph
from ._factory import CsrIndexedGraphFactory


@pytest.fixture
def edges() -> typing.Sequence[typing.Tuple[hpotk.TermId, hpotk.TermId]]:
    # Edges for a graph with a single root node: HP:1
    return (
        # source -> destination
        (hpotk.TermId.from_curie('HP:01'), hpotk.TermId.from_curie('HP:1')),
        (hpotk.TermId.from_curie('HP:010'), hpotk.TermId.from_curie('HP:01')),
        (hpotk.TermId.from_curie('HP:011'), hpotk.TermId.from_curie('HP:01')),
        # This one is multi-parent.
        (hpotk.TermId.from_curie('HP:0110'), hpotk.TermId.from_curie('HP:010')),
        (hpotk.TermId.from_curie('HP:0110'), hpotk.TermId.from_curie('HP:011')),

        (hpotk.TermId.from_curie('HP:02'), hpotk.TermId.from_curie('HP:1')),
        (hpotk.TermId.from_curie('HP:020'), hpotk.TermId.from_curie('HP:02')),
        (hpotk.TermId.from_curie('HP:021'), hpotk.TermId.from_curie('HP:02')),
        (hpotk.TermId.from_curie('HP:022'), hpotk.TermId.from_curie('HP:02')),

        (hpotk.TermId.from_curie('HP:03'), hpotk.TermId.from_curie('HP:1')),
    )


class TestCsrIndexedOntologyGraph:
    """
    Assuming `CsrIndexedGraphFactory` is correct, here we test the HPO traversal and others.
    """

    @pytest.fixture
    def graph(self, edges: typing.Sequence[typing.Tuple[hpotk.TermId, hpotk.TermId]]) -> CsrIndexedOntologyGraph:
        factory = CsrIndexedGraphFactory()
        return factory.create_graph(edges)

    def test_root(self, graph: CsrIndexedOntologyGraph):
        assert graph.root == hpotk.TermId.from_curie('HP:1')

    @pytest.mark.parametrize(
        'source, expected',
        [
            ("HP:1", {"HP:01", "HP:02", "HP:03"}),
            ("HP:01", {"HP:010", "HP:011"}),
            ("HP:010", {"HP:0110"}),
            ("HP:011", {"HP:0110"}),
            ("HP:0110", {}),

            ("HP:02", {"HP:020", "HP:021", "HP:022"}),

            ("HP:03", {}),
        ]
    )
    def test_get_children(self, graph: CsrIndexedOntologyGraph,
                          source: str, expected: typing.Set[str]):
        actual = set(graph.get_children(hpotk.TermId.from_curie(source)))

        assert actual == set(hpotk.TermId.from_curie(curie) for curie in expected)

    @pytest.mark.parametrize(
        'source, expected',
        [
            ("HP:1", {"HP:01", "HP:010", "HP:011", "HP:0110", "HP:02", "HP:020", "HP:021", "HP:022", "HP:03"}),
            ("HP:01", {"HP:010", "HP:011", "HP:0110"}),
            ("HP:010", {"HP:0110"}),
            ("HP:011", {"HP:0110"}),
            ("HP:0110", {}),

            ("HP:02", {"HP:020", "HP:021", "HP:022"}),

            ("HP:03", {}),
        ]
    )
    def test_get_descendants(self, graph: CsrIndexedOntologyGraph,
                             source: str, expected: typing.Set[str]):
        actual = set(graph.get_descendants(hpotk.TermId.from_curie(source)))

        assert actual == set(hpotk.TermId.from_curie(curie) for curie in expected)

    @pytest.mark.parametrize(
        'source, expected',
        [
            ("HP:1", {"HP:01", "HP:010", "HP:011", "HP:0110", "HP:02", "HP:020", "HP:021", "HP:022", "HP:03", "HP:1"}),
            ("HP:01", {"HP:010", "HP:011", "HP:0110", "HP:01"}),
            ("HP:010", {"HP:0110", "HP:010"}),
            ("HP:011", {"HP:0110", "HP:011"}),
            ("HP:0110", {"HP:0110"}),

            ("HP:02", {"HP:020", "HP:021", "HP:022", "HP:02"}),

            ("HP:03", {"HP:03"}),
        ]
    )
    def test_get_descendants_incl_source(self, graph: CsrIndexedOntologyGraph,
                                         source: str, expected: typing.Set[str]):
        actual = set(graph.get_descendants(hpotk.TermId.from_curie(source), include_source=True))

        assert actual == set(hpotk.TermId.from_curie(curie) for curie in expected)

    @pytest.mark.parametrize(
        'source, expected',
        [
            ("HP:1", {}),
            ("HP:01", {"HP:1"}),
            ("HP:010", {"HP:01"}),
            ("HP:0110", {"HP:010", "HP:011"}),

            ("HP:03", {"HP:1"}),
        ]
    )
    def test_get_parents(self, graph: CsrIndexedOntologyGraph,
                         source: str, expected: typing.Set[str]):
        actual = set(graph.get_parents(hpotk.TermId.from_curie(source)))

        assert actual == set(hpotk.TermId.from_curie(curie) for curie in expected)

    @pytest.mark.parametrize(
        'source, expected',
        [
            ("HP:1", {}),
            ("HP:01", {"HP:1"}),
            ("HP:010", {"HP:1", "HP:01"}),
            ("HP:0110", {"HP:1", "HP:01", "HP:010", "HP:011"}),

            ("HP:03", {"HP:1"}),
        ]
    )
    def test_get_ancestors(self, graph: CsrIndexedOntologyGraph,
                           source: str, expected: typing.Set[str]):
        actual = set(graph.get_ancestors(hpotk.TermId.from_curie(source)))

        assert actual == set(hpotk.TermId.from_curie(curie) for curie in expected)

    @pytest.mark.parametrize(
        'source, expected',
        [
            ("HP:1", {"HP:1"}),
            ("HP:01", {"HP:1", "HP:01"}),
            ("HP:010", {"HP:1", "HP:01", "HP:010"}),
            ("HP:0110", {"HP:1", "HP:01", "HP:010", "HP:011", "HP:0110"}),

            ("HP:03", {"HP:1", "HP:03"}),
        ]
    )
    def test_get_ancestors_incl_source(self, graph: CsrIndexedOntologyGraph,
                                       source: str, expected: typing.Set[str]):
        actual = set(graph.get_ancestors(hpotk.TermId.from_curie(source), include_source=True))

        assert actual == set(hpotk.TermId.from_curie(curie) for curie in expected)



import unittest

from collections import namedtuple

import ddt

from ._csr import *

CsrData = namedtuple('CsrData', ['indptr', 'indices', 'data', 'shape'])


@ddt.ddt
class TestImmutableCsrMatrix(unittest.TestCase):

    # [[100 102 104 106]
    #  [108 110 112 114]
    #  [116 118 120 122]
    #  [124 126 128 130]]
    FULL = CsrData(
        indptr=[0, 4, 8, 12, 16],
        indices=[0, 1, 2, 3, 0, 1, 2, 3, 0, 1, 2, 3, 0, 1, 2, 3],
        data=[100, 102, 104, 106, 108, 110, 112, 114, 116, 118, 120, 122, 124, 126, 128, 130],
        shape=(4, 4))

    # [[1. 0. 0. 2.]
    #  [0. 3. 0. 0.]
    #  [0. 0. 4. 0.]
    #  [5. 0. 0. 6.]]
    ALL_EDGES = CsrData(
        indptr=[0, 2, 3, 4, 6],
        indices=[0, 3, 1, 2, 0, 3],
        data=[1., 2., 3., 4., 5., 6.],
        shape=(4, 4))

    # [[0. 0. 0.]
    #  [0. 0. 0.]
    #  [0. 0. 0.]]
    ZEROES = CsrData(
        indptr=[0, 0, 0, 0],
        indices=[],
        data=[],
        shape=(3, 3))

    # [[0. 1. 0. 2.]
    #  [3. 0. 4. 5.]]
    RECT = CsrData(
        indptr=[0, 2, 5],
        indices=[1, 3, 0, 2, 3],
        data=[1., 2., 3., 4., 5.],
        shape=(2, 4))

    @ddt.data(
        (0, 0, 100), (0, 1, 102), (0, 2, 104), (0, 3, 106),
        (1, 0, 108), (1, 1, 110), (1, 2, 112), (1, 3, 114),
        (2, 0, 116), (2, 1, 118), (2, 2, 120), (2, 3, 122),
        (3, 0, 124), (3, 1, 126), (3, 2, 128), (3, 3, 130))
    @ddt.unpack
    def test_full(self, row, col, val):
        mat = make_csr_matrix(TestImmutableCsrMatrix.FULL)
        self.assertEqual(val, mat[row, col])

    @ddt.data(
        (0, 0, 1.), (0, 1, 0.), (0, 2, 0.), (0, 3, 2.),
        (1, 0, 0.), (1, 1, 3.), (1, 2, 0.), (1, 3, 0.),
        (2, 0, 0.), (2, 1, 0.), (2, 2, 4.), (2, 3, 0.),
        (3, 0, 5.), (3, 1, 0.), (3, 2, 0.), (3, 3, 6.))
    @ddt.unpack
    def test_all_edges(self, row, col, val):
        mat = make_csr_matrix(TestImmutableCsrMatrix.ALL_EDGES)
        self.assertEqual(val, mat[row, col])

    def test_zeros(self):
        zeros = make_csr_matrix(TestImmutableCsrMatrix.ZEROES)
        for i in range(zeros.shape[0]):
            for j in range(zeros.shape[1]):
                self.assertEqual(0., zeros[i, j])

    def test_shapes(self):
        self.assertEqual((4, 4), make_csr_matrix(TestImmutableCsrMatrix.FULL).shape)
        self.assertEqual((4, 4), make_csr_matrix(TestImmutableCsrMatrix.ALL_EDGES).shape)
        self.assertEqual((3, 3), make_csr_matrix(TestImmutableCsrMatrix.ZEROES).shape)
        self.assertEqual((2, 4), make_csr_matrix(TestImmutableCsrMatrix.RECT).shape)

    @ddt.data(
        (0, [100, 102, 104, 106]),
        (1, [108, 110, 112, 114]),
        (2, [116, 118, 120, 122]),
        (3, [124, 126, 128, 130]))
    @ddt.unpack
    def test_get_row__full(self, i, vals):
        mat = make_csr_matrix(TestImmutableCsrMatrix.FULL)
        assert np.allclose(mat[i], np.array(vals))

    @ddt.data(
        (0, [0., 0., 0.]),
        (1, [0., 0., 0.]),
        (2, [0., 0., 0.]))
    @ddt.unpack
    def test_get_row__zeroes(self, i, vals):
        mat = make_csr_matrix(TestImmutableCsrMatrix.ZEROES)
        assert np.allclose(mat[i], np.array(vals))

    @ddt.data(
        (0, [1., 0., 0., 2.]),
        (1, [0., 3., 0., 0.]),
        (2, [0., 0., 4., 0.]),
        (3, [5., 0., 0., 6.]))
    @ddt.unpack
    def test_get_row__all_edges(self, i, vals):
        mat = make_csr_matrix(TestImmutableCsrMatrix.ALL_EDGES)
        assert np.allclose(mat[i], np.array(vals))

    def test_col_indices_of_val(self):
        edges = make_csr_matrix(TestImmutableCsrMatrix.ALL_EDGES)
        assert np.allclose(edges.col_indices_of_val(0, 0.), [1, 2])
        assert np.allclose(edges.col_indices_of_val(0, 1.), [0])
        assert np.allclose(edges.col_indices_of_val(0, 2.), [3])
        assert np.allclose(edges.col_indices_of_val(0, 3.), [])

        zeroes = make_csr_matrix(TestImmutableCsrMatrix.ZEROES)
        assert np.allclose(zeroes.col_indices_of_val(0, 0.), [0, 1, 2])
        assert np.allclose(zeroes.col_indices_of_val(0, 1.), [])

        full = make_csr_matrix(TestImmutableCsrMatrix.FULL)
        assert np.allclose(full.col_indices_of_val(0, 0), [])
        assert np.allclose(full.col_indices_of_val(0, 100), [0])
        assert np.allclose(full.col_indices_of_val(0, 102), [1])


def make_csr_matrix(example: CsrData):
    return ImmutableCsrMatrix(example.indptr, example.indices, example.data, example.shape)


class TestCsrMatrixBuilder(unittest.TestCase):

    def test_incremental(self):
        builder = CsrMatrixBuilder(shape=(3, 3))
        builder[0, 0] = 1.
        builder[0, 1] = 2.
        builder[0, 2] = 3.

        builder[1, 0] = 4.
        builder[1, 1] = 5.
        builder[1, 2] = 6.

        builder[2, 0] = 7.
        builder[2, 1] = 8.
        builder[2, 2] = 9.

        assert np.allclose(builder.row, np.array([0, 3, 6, 9]))
        assert np.allclose(builder.col, np.array([0, 1, 2, 0, 1, 2, 0, 1, 2]))
        assert np.allclose(builder.data, np.array([1, 2, 3, 4, 5, 6, 7, 8, 9]))

    def test_decrement(self):
        builder = CsrMatrixBuilder(shape=(3, 3))
        builder[2, 2] = 9.
        builder[2, 1] = 8.
        builder[2, 0] = 7.

        builder[1, 2] = 6.
        builder[1, 1] = 5.
        builder[1, 0] = 4.

        builder[0, 2] = 3.
        builder[0, 1] = 2.
        builder[0, 0] = 1.

        assert np.allclose(builder.row, np.array([0, 3, 6, 9]))
        assert np.allclose(builder.col, np.array([0, 1, 2, 0, 1, 2, 0, 1, 2]))
        assert np.allclose(builder.data, np.array([1, 2, 3, 4, 5, 6, 7, 8, 9]))



import abc
import itertools
import typing

from hpotk.model import TermId, Identified

# TODO - enforce presence of the natural ordering?
# Note, the NODE must also have natural ordering.
NODE = typing.TypeVar('NODE', bound=TermId)
# Term ID that is added as an artificial root if >1 root candidates are found in the ontology graph.
OWL_THING = TermId.from_curie("owl:Thing")


class OntologyGraph(typing.Generic[NODE], metaclass=abc.ABCMeta):
    """
    A simple graph with one node type and one edge type.

    The graph is generic over a node type which must extend :class:`TermId`.
    The graph must not be empty, it must consist of at least one node.

    .. note::

      `OntologyGraph` provides **iterators** for traversals instead of sets, lists, etc.
      See :ref:`iterable-vs-iterator` to learn why.
    """

    @property
    @abc.abstractmethod
    def root(self) -> NODE:
        """
        Get the root node of the ontology graph.
        """
        pass

    @abc.abstractmethod
    def get_children(self, source: typing.Union[str, NODE, Identified],
                     include_source: bool = False) -> typing.Iterator[NODE]:
        """
        Get an iterator with the children of the `source` node.

        :param source: a :class:`TermId`, an item that *has* a :class:`TermId` (:class:`Identified`), or a curie `str`
          representing the source node.
        :param include_source: `True` if the `source` should be included among the children, `False` otherwise.
        :raises ValueError: if `source` is not present in the graph.
        """
        pass

    @abc.abstractmethod
    def get_descendants(self, source: typing.Union[str, NODE, Identified],
                        include_source: bool = False) -> typing.Iterator[NODE]:
        """
        Get an iterator with the descendants of the `source` node.

        :param source: a :class:`TermId`, an item that *has* a :class:`TermId` (:class:`Identified`), or a curie `str`
          representing the source node.
        :param include_source: `True` if the `source` should be included among the descendants, `False` otherwise.
        :raises ValueError: if `source` is not present in the graph.
        """
        pass

    @abc.abstractmethod
    def get_parents(self, source: typing.Union[str, NODE, Identified],
                    include_source: bool = False) -> typing.Iterator[NODE]:
        """
        Get an iterator with the parents of the `source` node.

        :param source: a :class:`TermId`, an item that *has* a :class:`TermId` (:class:`Identified`), or a curie `str`
          representing the source node.
        :param include_source: `True` if the `source` should be included among the parents, `False` otherwise.
        :raises ValueError: if `source` is not present in the graph.
        """
        pass

    @abc.abstractmethod
    def get_ancestors(self, source: typing.Union[str, NODE, Identified],
                      include_source: bool = False) -> typing.Iterator[NODE]:
        """
        Get an iterator with the ancestors of the `source` node.

        :param source: a :class:`TermId`, an item that *has* a :class:`TermId` (:class:`Identified`), or a curie `str`
          representing the source node.
        :param include_source: `True` if the `source` should be included among the ancestors, `False` otherwise.
        :raises ValueError: if `source` is not present in the graph.
        """
        pass

    def is_leaf(self, node: typing.Union[str, NODE, Identified]) -> bool:
        """
        Test if the node is a leaf - a node with no children.

        :return: `True` if the `node` is a leaf node or `False` otherwise.
        :raises ValueError: if `node` is not present in the graph.
        """
        for _ in self.get_children(node):
            return False
        return True

    def is_parent_of(self, sub: typing.Union[str, NODE, Identified],
                     obj: typing.Union[str, NODE, Identified]) -> bool:
        """
        Return `True` if the subject `sub` is a parent of the object `obj`.

        :param sub: a graph node.
        :param obj: other graph node.
        :return: `True` if the `sub` is a parent of the `obj`.
        :raises ValueError: if `obj` is not present in the graph.
        """
        return self._run_query(self.get_parents, sub, obj)

    def is_ancestor_of(self, sub: typing.Union[str, NODE, Identified],
                       obj: typing.Union[str, NODE, Identified]) -> bool:
        """
        Return `True` if the subject `sub` is an ancestor of the object `obj`.

        :param sub: a graph node.
        :param obj: other graph node.
        :return: `True` if the `sub` is an ancestor of the `obj`.
        :raises ValueError: if `obj` is not present in the graph.
        """
        return self._run_query(self.get_ancestors, sub, obj)

    def is_child_of(self, sub: typing.Union[str, NODE, Identified],
                    obj: typing.Union[str, NODE, Identified]) -> bool:
        """
        Return `True` if the `sub` is a child of the `obj`.

        :param sub: a graph node.
        :param obj: other graph node.
        :return: `True` if the `sub` is a child of the `obj`.
        :raises ValueError: if `obj` is not present in the graph.
        """
        return self._run_query(self.get_children, sub, obj)

    def is_descendant_of(self, sub: typing.Union[str, NODE, Identified],
                         obj: typing.Union[str, NODE, Identified]) -> bool:
        """
        Return `True` if the `sub` is a descendant of the `obj`.

        :param sub: a graph node.
        :param obj: other graph node.
        :return: `True` if the `sub` is a descendant of the `obj`.
        :raises ValueError: if `obj` is not present in the graph.
        """
        return self._run_query(self.get_descendants, sub, obj)

    @staticmethod
    def _run_query(func: typing.Callable[[NODE], typing.Iterator[NODE]],
                   sub: typing.Union[str, NODE, Identified],
                   obj: typing.Union[str, NODE, Identified]) -> bool:
        sub = OntologyGraph._map_to_term_id(sub)
        obj = OntologyGraph._map_to_term_id(obj)
        return any(sub == term_id for term_id in func(obj))

    @abc.abstractmethod
    def __contains__(self, item: NODE) -> bool:
        pass

    @abc.abstractmethod
    def __iter__(self) -> typing.Iterator[NODE]:
        pass

    @staticmethod
    def _map_to_term_id(item: typing.Union[str, NODE, Identified]) -> TermId:
        if isinstance(item, TermId):
            return item
        elif isinstance(item, Identified):
            return item.identifier
        elif isinstance(item, str):
            return TermId.from_curie(item)
        else:
            raise ValueError(f'Expected `TermId`, `Identified`, or `str` but got `{type(item)}`')


class IndexedOntologyGraph(typing.Generic[NODE], OntologyGraph[NODE], metaclass=abc.ABCMeta):
    """
    `IndexedOntologyGraph` allows working with ontology graph node indices instead of the ontology graph nodes.
    Working in the index space is generally faster, when used to traverse the graph or to create term id unions,
    differences, etc...

    Starting from a node index, `IndexedOntologyGraph` provides methods for getting indices of its children, descendants,
    parents, and ancestors. The node index can be obtained from :func:`node_to_idx`. Having an index, you can get
    the corresponding node using :func:`idx_to_node`.
    """

    @property
    @abc.abstractmethod
    def root_idx(self) -> int:
        """
        Get the index of the root node of the ontology graph.
        """
        pass

    @abc.abstractmethod
    def get_children_idx(self, source: int) -> typing.Sequence[int]:
        """
        Get an iterator with the indices of the children of the `source` node.

        :param source: an index of a node that represents the source node.
        :raises ValueError: if `source` is not present in the graph.
        """
        pass

    @abc.abstractmethod
    def get_descendant_idx(self, source: int) -> typing.Iterator[int]:
        """
        Get an iterator with the indices of the descendants of the `source` node.

        :param source: an index of a node that represents the source node.
        :raises ValueError: if `source` is not present in the graph.
        """
        pass

    @abc.abstractmethod
    def get_parents_idx(self, source: int) -> typing.Sequence[int]:
        """
        Get an iterator with the indices of the parents of the `source` node.

        :param source: an index of a node that represents the source node.
        :raises ValueError: if `source` is not present in the graph.
        """
        pass

    @abc.abstractmethod
    def get_ancestor_idx(self, source: int) -> typing.Iterator[int]:
        """
        Get an iterator with the indices of the ancestors of the `source` node.

        :param source: an index of a node that represents the source node.
        :raises ValueError: if `source` is not present in the graph.
        """
        pass

    @abc.abstractmethod
    def idx_to_node(self, idx: int) -> NODE:
        """
        Map the index into the corresponding node.

        :param idx: index to map to a node.
        :return: the node corresponding to the index.
        :raises ValueError: if `idx` does not correspond to any nodes of the ontology graph.
        """
        pass

    @abc.abstractmethod
    def node_to_idx(self, node: NODE) -> typing.Optional[int]:
        """
        Map the node into the corresponding node index.

        :param node: node to retrieve an index for.
        :return: the index corresponding to the `node` or `None` if the `node` is not in the graph.
        """
        pass

    # Override the `OntologyGraph` parts  ############################################################################ #

    @property
    def root(self) -> NODE:
        return self.idx_to_node(self.root_idx)

    def get_children(self, source: typing.Union[str, NODE, Identified],
                     include_source: bool = False) -> typing.Iterator[NODE]:
        return self._map_with_seq_func(source, include_source, self.get_children_idx)

    def get_descendants(self, source: typing.Union[str, NODE, Identified],
                        include_source: bool = False) -> typing.Iterator[NODE]:
        return self._map_with_iter_func(source, include_source, self.get_descendant_idx)

    def get_parents(self, source: typing.Union[str, NODE, Identified],
                    include_source: bool = False) -> typing.Iterator[NODE]:
        return self._map_with_seq_func(source, include_source, self.get_parents_idx)

    def get_ancestors(self, source: typing.Union[str, NODE, Identified],
                      include_source: bool = False) -> typing.Iterator[NODE]:
        return self._map_with_iter_func(source, include_source, self.get_ancestor_idx)

    def is_leaf(self, node: typing.Union[str, NODE, Identified]) -> bool:
        node_idx = self._map_to_term_idx(node)
        if node_idx is None:
            raise ValueError(f'No graph node found for {node}')

        for _ in self.get_children_idx(node_idx):
            return False
        return True

    def is_parent_of_idx(self, sub: int, obj: int) -> bool:
        """
        Return `True` if the subject `sub` is a parent of the object `obj`.

        :param sub: index of a graph node.
        :param obj: index of the other graph node.
        :return: `True` if the `sub` is a parent of the `obj`.
        :raises ValueError: if no such node exists for the `obj` index.
        """
        return any(sub == idx for idx in self.get_parents_idx(obj))

    def is_parent_of(self, sub: typing.Union[str, NODE, Identified],
                     obj: typing.Union[str, NODE, Identified]) -> bool:
        obj_idx = self._map_to_term_idx(obj)
        if obj_idx is None:
            raise ValueError(f'No graph node found for {obj}')

        sub_idx = self._map_to_term_idx(sub)
        if sub_idx is None:
            return False

        return any(sub_idx == idx for idx in self.get_parents_idx(obj_idx))

    def is_ancestor_of_idx(self, sub: int, obj: int) -> bool:
        """
        Return `True` if the subject `sub` is an ancestor of the object `obj`.

        :param sub: index of a graph node.
        :param obj: index of the other graph node.
        :return: `True` if the `sub` is an ancestor of the `obj`.
        :raises ValueError: if no such node exists for the `obj` index.
        """
        return any(sub == idx for idx in self.get_ancestor_idx(obj))

    def is_ancestor_of(self, sub: typing.Union[str, NODE, Identified],
                       obj: typing.Union[str, NODE, Identified]) -> bool:
        obj_idx = self._map_to_term_idx(obj)
        if obj_idx is None:
            raise ValueError(f'No graph node found for {obj}')

        sub_idx = self._map_to_term_idx(sub)
        if sub_idx is None:
            return False

        return any(sub_idx == idx for idx in self.get_ancestor_idx(obj_idx))

    def is_child_of_idx(self, sub: int, obj: int) -> bool:
        """
        Return `True` if the subject `sub` is a child of the object `obj`.

        :param sub: index of a graph node.
        :param obj: index of the other graph node.
        :return: `True` if the `sub` is a child of the `obj`.
        :raises ValueError: if no such node exists for the `sub` index.
        """
        # TODO: ValueError for `sub` may break the pattern
        return any(obj == idx for idx in self.get_parents_idx(sub))

    def is_child_of(self, sub: typing.Union[str, NODE, Identified],
                    obj: typing.Union[str, NODE, Identified]) -> bool:
        obj_idx = self._map_to_term_idx(obj)
        if obj_idx is None:
            raise ValueError(f'No graph node found for {obj}')

        sub_idx = self._map_to_term_idx(sub)
        if sub_idx is None:
            return False

        # Exploit the fact that a term has usually fewer parents than children.
        return any(obj_idx == idx for idx in self.get_parents_idx(sub_idx))

    def is_descendant_of_idx(self, sub: int, obj: int) -> bool:
        """
        Return `True` if the subject `sub` is a descendant of the object `obj`.

        :param sub: index of a graph node.
        :param obj: index of the other graph node.
        :return: `True` if the `sub` is a descendant of the `obj`.
        :raises ValueError: if no such node exists for the `sub` index.
        """
        # TODO: ValueError for `sub` may break the pattern
        return any(obj == idx for idx in self.get_ancestor_idx(sub))

    def is_descendant_of(self, sub: typing.Union[str, NODE, Identified],
                         obj: typing.Union[str, NODE, Identified]) -> bool:
        obj_idx = self._map_to_term_idx(obj)
        if obj_idx is None:
            raise ValueError(f'No graph node found for {obj}')

        sub_idx = self._map_to_term_idx(sub)
        if sub_idx is None:
            return False

        # Exploit the fact that a term has usually fewer parents than children.
        return any(obj_idx == idx for idx in self.get_ancestor_idx(sub_idx))

    def _map_with_iter_func(self, node: typing.Union[str, NODE, Identified],
                            include_source: bool,
                            func: typing.Callable[[int], typing.Iterator[int]]) -> typing.Iterator[NODE]:
        idx = self._map_to_term_idx(node)
        if idx is not None:
            if include_source:
                return itertools.chain(
                    (self.idx_to_node(idx),),
                    map(self.idx_to_node, func(idx)))
            else:
                return map(self.idx_to_node, func(idx))
        else:
            raise ValueError(f'{node} is not present in the graph!')

    def _map_with_seq_func(self, node: typing.Union[str, NODE, Identified],
                           include_source: bool,
                           func: typing.Callable[[int], typing.Sequence[int]]) -> typing.Iterator[NODE]:
        idx = self._map_to_term_idx(node)
        if idx is not None:
            if include_source:
                return itertools.chain(
                    (self.idx_to_node(idx),),
                    map(self.idx_to_node, func(idx)))
            else:
                return map(self.idx_to_node, func(idx))
        else:
            raise ValueError(f'{node} is not present in the graph!')

    def _map_to_term_idx(self, node: typing.Union[str, NODE, Identified]) -> typing.Optional[int]:
        """
        A convenience method to convert a `node` into the node index.

        :param node: one of the expected node types, including CURIE `str`, NODE, or an :class:`Identified` item.
        :return: the node index or `None` if the node is not present in the graph.
        :raises ValueError: if the node is not in one of the expected types.
        """
        term_id = self._map_to_term_id(node)
        return self.node_to_idx(term_id)

    # The rest

    def __contains__(self, item: NODE) -> bool:
        return self.node_to_idx(item) is not None


class GraphAware(typing.Generic[NODE], metaclass=abc.ABCMeta):
    """
    A mixin class for entities that have an :class:`OntologyGraph`.
    """

    @property
    @abc.abstractmethod
    def graph(self) -> OntologyGraph[NODE]:
        """
        Get the ontology graph.
        """
        pass



from hpotk.model import TermId

# Children of Phenotypic abnormality `HP:0000118`.

ABNORMALITY_OF_BLOOD_AND_BLOOD_FORMING_TISSUES: TermId = TermId.from_curie("HP:0001871")
ABNORMALITY_OF_GENITOURINARY_SYSTEM: TermId = TermId.from_curie("HP:0000119")
ABNORMALITY_OF_HEAD_OR_NECK: TermId = TermId.from_curie("HP:0000152")
ABNORMALITY_OF_METABOLISM_OR_HOMEOSTASIS: TermId = TermId.from_curie("HP:0001939")
ABNORMALITY_OF_PRENATAL_DEVELOPMENT_OR_BIRTH: TermId = TermId.from_curie("HP:0001197")
ABNORMALITY_OF_THE_BREAST: TermId = TermId.from_curie("HP:0000769")
ABNORMALITY_OF_THE_CARDIOVASCULAR_SYSTEM: TermId = TermId.from_curie("HP:0001626")
ABNORMALITY_OF_THE_DIGESTIVE_SYSTEM: TermId = TermId.from_curie("HP:0025031")
ABNORMALITY_OF_THE_EAR: TermId = TermId.from_curie("HP:0000598")
ABNORMALITY_OF_THE_ENDOCRINE_SYSTEM: TermId = TermId.from_curie("HP:0000818")
ABNORMALITY_OF_THE_EYE: TermId = TermId.from_curie("HP:0000478")
ABNORMALITY_OF_THE_IMMUNE_SYSTEM: TermId = TermId.from_curie("HP:0002715")
ABNORMALITY_OF_THE_INTEGUMENT: TermId = TermId.from_curie("HP:0001574")
ABNORMALITY_OF_THE_LIMBS: TermId = TermId.from_curie("HP:0040064")
ABNORMALITY_OF_THE_MUSCULOSKELETAL_SYSTEM: TermId = TermId.from_curie("HP:0033127")
ABNORMALITY_OF_THE_NERVOUS_SYSTEM: TermId = TermId.from_curie("HP:0000707")
ABNORMALITY_OF_THE_RESPIRATORY_SYSTEM: TermId = TermId.from_curie("HP:0002086")
ABNORMALITY_OF_THE_THORACIC_CAVITY: TermId = TermId.from_curie("HP:0045027")
ABNORMALITY_OF_THE_VOICE: TermId = TermId.from_curie("HP:0001608")
ABNORMAL_CELLULAR_PHENOTYPE: TermId = TermId.from_curie("HP:0025354")
CONSTITUTIONAL_SYMPTOM: TermId = TermId.from_curie("HP:0025142")
GROWTH_ABNORMALITY: TermId = TermId.from_curie("HP:0001507")
NEOPLASM: TermId = TermId.from_curie("HP:0002664")


# These two are in phenol, but they are located under musculoskeletal system
# ABNORMALITY_OF_THE_CONNECTIVE_TISSUE: TermId = TermId.from_curie("HP:0003549")
# SKELETAL_ID: TermId = TermId.from_curie("HP:0000924")



from ._load import load_ontology, load_minimal_ontology
from ._factory import TermFactory, MinimalTermFactory

__all__ = [
    'load_ontology', 'load_minimal_ontology',
    'TermFactory', 'MinimalTermFactory'
]



import unittest

import ddt
import numpy as np

from hpotk.model import TermId

from .csr import ImmutableCsrMatrix
from ._csr_graph import check_items_are_sorted
from ._csr_graph import BisectPoweredCsrOntologyGraph


@ddt.ddt
class TestFunctions(unittest.TestCase):

    @ddt.data(
        ([],),
        ([1],),
        ([1, 2],),
        (['a', 'b', 'c'],),
    )
    @ddt.unpack
    def test_check_items_are_sorted(self, items):
        self.assertIsNone(check_items_are_sorted(items))

    @ddt.data(
        ([2, 1], 'Unsorted sequence. Item #1 (1) was less than #0 (2)'),
        ([2, 1, 3, 4, 5, 6], 'Unsorted sequence. Item #1 (1) was less than #0 (2)'),
        ([1, 2, 30, 4, 5, 6], 'Unsorted sequence. Item #3 (4) was less than #2 (30)'),
        ([1, 2, 3, 4, 6, 5], 'Unsorted sequence. Item #5 (5) was less than #4 (6)'),
    )
    @ddt.unpack
    def test_check_items_are_sorted__unsorted_cases(self, items, message):
        with self.assertRaises(ValueError) as ctx:
            check_items_are_sorted(items)
        self.assertEqual(message, ctx.exception.args[0])


@ddt.ddt
class TestBisectPoweredCsrOntologyGraph(unittest.TestCase):

    NODES: np.ndarray
    GRAPH: BisectPoweredCsrOntologyGraph

    @classmethod
    def setUpClass(cls) -> None:
        root = TermId.from_curie('HP:1')
        curies = [
            'HP:01', 'HP:010', 'HP:011', 'HP:0110',
            'HP:02', 'HP:020', 'HP:021', 'HP:022',
            'HP:03', 'HP:1'
        ]
        nodes = np.fromiter(map(TermId.from_curie, curies), dtype=object)
        row = [0, 3, 5, 7, 9, 13, 14, 15, 16, 17, 20]
        col = [1, 2, 9, 0, 3, 0, 3, 1, 2, 5, 6, 7, 9, 4, 4, 4, 9, 0, 4, 8]
        data = [-1, -1, 1, 1, -1, 1, -1, 1, 1, -1, -1, -1, 1, 1, 1, 1, 1, -1, -1, -1]
        am = ImmutableCsrMatrix(row, col, data, shape=(len(nodes), len(nodes)), dtype=int)

        cls.NODES = nodes
        cls.GRAPH = BisectPoweredCsrOntologyGraph(root, nodes, am)

    def test_root(self):
        self.assertEqual(TermId.from_curie('HP:1'), self.GRAPH.root)

    @ddt.data(
        ('HP:1', ['HP:01', 'HP:02', 'HP:03']),
        ('HP:02', ['HP:020', 'HP:021', 'HP:022']),
        ('HP:03', []),
    )
    @ddt.unpack
    def test_get_children(self, source, expected):
        src = TermId.from_curie(source)
        exp = set(map(TermId.from_curie, expected))

        for term_id in self.GRAPH.get_children(src):
            exp.remove(term_id)
        self.assertTrue(len(exp) == 0)

        exp = set(map(TermId.from_curie, expected))
        exp.add(src)
        for term_id in self.GRAPH.get_children(src, include_source=True):
            exp.remove(term_id)

        self.assertTrue(len(exp) == 0)

    @ddt.data(
        ('HP:1', ['HP:01', 'HP:010', 'HP:011', 'HP:0110',
                  'HP:02', 'HP:020', 'HP:021', 'HP:022',
                  'HP:03']),
        ('HP:01', ['HP:010', 'HP:011', 'HP:0110']),
        ('HP:010', ['HP:0110']),
        ('HP:011', ['HP:0110']),
        ('HP:0110', []),

        ('HP:02', ['HP:020', 'HP:021', 'HP:022']),
        ('HP:020', []),
        ('HP:021', []),
        ('HP:022', []),
        ('HP:03', []),
    )
    @ddt.unpack
    def test_get_descendants(self, source, expected):
        src = TermId.from_curie(source)
        exp = set(map(TermId.from_curie, expected))

        for term_id in self.GRAPH.get_descendants(src):
            exp.remove(term_id)
        self.assertTrue(len(exp) == 0)

        exp = set(map(TermId.from_curie, expected))
        exp.add(src)
        for term_id in self.GRAPH.get_descendants(src, include_source=True):
            exp.remove(term_id)

        self.assertTrue(len(exp) == 0)

    @ddt.data(
        ('HP:1', []),
        ('HP:01', ['HP:1']),
        ('HP:03', ['HP:1']),
        ('HP:0110', ['HP:010', 'HP:011']),
    )
    @ddt.unpack
    def test_get_parents(self, source, expected):
        src = TermId.from_curie(source)
        exp = set(map(TermId.from_curie, expected))

        for term_id in self.GRAPH.get_parents(src):
            exp.remove(term_id)
        self.assertTrue(len(exp) == 0)

        exp = set(map(TermId.from_curie, expected))
        exp.add(src)
        for term_id in self.GRAPH.get_parents(src, include_source=True):
            exp.remove(term_id)

        self.assertTrue(len(exp) == 0)

    @ddt.data(
        ('HP:1', []),
        ('HP:01', ['HP:1']),
        ('HP:0110', ['HP:010', 'HP:011', 'HP:01', 'HP:1']),
        ('HP:022', ['HP:02', 'HP:1']),
        ('HP:03', ['HP:1']),
    )
    @ddt.unpack
    def test_get_ancestors(self, source, expected):
        src = TermId.from_curie(source)
        exp = set(map(TermId.from_curie, expected))

        for term_id in self.GRAPH.get_ancestors(src):
            exp.remove(term_id)
        self.assertTrue(len(exp) == 0)

        exp = set(map(TermId.from_curie, expected))
        exp.add(src)
        for term_id in self.GRAPH.get_ancestors(src, include_source=True):
            exp.remove(term_id)

        self.assertTrue(len(exp) == 0)

    @ddt.data(
        ('HP:1', False),

        ('HP:01', False),
        ('HP:010', False),
        ('HP:011', False),
        ('HP:0110', True),

        ('HP:02', False),
        ('HP:020', True),
        ('HP:021', True),
        ('HP:022', True),

        ('HP:03', True),
    )
    @ddt.unpack
    def test_is_leaf(self, source, expected):
        src = TermId.from_curie(source)
        actual = self.GRAPH.is_leaf(src)
        self.assertEqual(expected, actual)

    def test_is_leaf__unknown_source(self):
        with self.assertRaises(ValueError) as ctx:
            self.GRAPH.is_leaf(TermId.from_curie('HP:999'))
        self.assertEqual('Term ID not found in the graph: HP:999', ctx.exception.args[0])

    @ddt.data(
        ('get_parents',),
        ('get_ancestors',),
        ('get_children',),
        ('get_descendants',),
    )
    @ddt.unpack
    def test_traverse_methods_with_unknown_source(self, func_name):
        unknown = TermId.from_curie('HP:999')

        func = getattr(self.GRAPH, func_name)
        with self.assertRaises(ValueError) as ctx:
            list(func(unknown))  # We must consume the iterable!
        self.assertEqual('Term ID not found in the graph: HP:999', ctx.exception.args[0])

        func = getattr(self.GRAPH, func_name)
        with self.assertRaises(ValueError) as ctx:
            list(func(unknown, include_source=True))  # We must consume the iterable!
        self.assertEqual('Term ID not found in the graph: HP:999', ctx.exception.args[0])

    def test_contains(self):
        for node in self.NODES:
            self.assertTrue(node in self.GRAPH)

    def test_iter(self):
        expected = set(self.NODES)
        actual = set(iter(self.GRAPH))
        self.assertSetEqual(expected, actual)




import unittest
import ddt

from ._impl import Ratio


@ddt.ddt
class TestRatio(unittest.TestCase):

    def test_positive(self):
        ratio = Ratio(1, 4)
        self.assertEqual(ratio.numerator, 1)
        self.assertEqual(ratio.denominator, 4)
        self.assertAlmostEqual(ratio.frequency, 1/4)
        self.assertTrue(ratio.is_positive())
        self.assertFalse(ratio.is_zero())

    def test_zero(self):
        ratio = Ratio(0, 1)
        self.assertEqual(ratio.numerator, 0)
        self.assertEqual(ratio.denominator, 1)
        self.assertAlmostEqual(ratio.frequency, 0.)
        self.assertFalse(ratio.is_positive())
        self.assertTrue(ratio.is_zero())

    @ddt.unpack
    @ddt.data(
        [1, 1, 2, 2, True],
        [1, 4, 2, 8, True],
        [2, 8, 1, 4, True],

        [1, 4, 3, 8, False],
        [1, 8, 1, 4, False],
    )
    def test_equality(self, left_num, left_denom, right_num, right_denom, expected):
        left = Ratio(left_num, left_denom)
        right = Ratio(right_num, right_denom)
        self.assertEqual(left == right, expected)

    @ddt.unpack
    @ddt.data(
        [1, 1, 2, 2, 3, 3],
        [1, 2, 3, 4, 4, 6],
    )
    def test_fold(self, left_num, left_denom, right_num, right_denom, result_num, result_denom):
        left = Ratio(left_num, left_denom)
        right = Ratio(right_num, right_denom)

        result = Ratio.fold(left, right)

        self.assertEqual(result.numerator, result_num)
        self.assertEqual(result.denominator, result_denom)




import typing

from hpotk.util import validate_instance, validate_optional_instance
from hpotk.graph import OntologyGraph
from hpotk.model import TermId, Identified
from ._api import Ontology, MinimalOntology
from ._api import ID, MINIMAL_TERM, TERM, CURIE_OR_TERM_ID_OR_IDENTIFIED


class DefaultMinimalOntology(MinimalOntology[ID, MINIMAL_TERM]):

    def __init__(self, graph: OntologyGraph[ID],
                 current_terms: typing.Sequence[MINIMAL_TERM],
                 term_id_to_term: typing.Mapping[ID, MINIMAL_TERM],
                 version: typing.Optional[str] = None):
        self._graph = validate_instance(graph, OntologyGraph, 'graph')
        self._current_terms = current_terms
        self._term_id_to_term = term_id_to_term
        self._version = validate_optional_instance(version, str, 'version')

    @property
    def graph(self) -> OntologyGraph[ID]:
        return self._graph

    @property
    def term_ids(self) -> typing.Iterator[ID]:
        return iter(self._term_id_to_term.keys())

    @property
    def terms(self) -> typing.Iterator[MINIMAL_TERM]:
        return iter(self._current_terms)

    def get_term(self, term_id: CURIE_OR_TERM_ID_OR_IDENTIFIED) -> typing.Optional[MINIMAL_TERM]:
        term_id = _validate_term_id(term_id)
        try:
            return self._term_id_to_term[term_id]
        except KeyError:
            return None

    @property
    def version(self) -> typing.Optional[str]:
        return self._version

    def __len__(self):
        return len(self._current_terms)


class DefaultOntology(Ontology[ID, TERM]):

    def __init__(self, graph: OntologyGraph[ID],
                 current_terms: typing.Sequence[TERM],
                 term_id_to_term: typing.Mapping[ID, TERM],
                 version: typing.Optional[str] = None):
        self._graph = validate_instance(graph, OntologyGraph, 'graph')
        self._current_terms = current_terms
        self._term_id_to_term = term_id_to_term
        self._version = validate_optional_instance(version, str, 'version')

    @property
    def graph(self) -> OntologyGraph[ID]:
        return self._graph

    @property
    def term_ids(self) -> typing.Iterator[ID]:
        return iter(self._term_id_to_term.keys())

    @property
    def terms(self) -> typing.Iterator[TERM]:
        return iter(self._current_terms)

    def get_term(self, term_id: CURIE_OR_TERM_ID_OR_IDENTIFIED) -> typing.Optional[TERM]:
        term_id = _validate_term_id(term_id)
        try:
            return self._term_id_to_term[term_id]
        except KeyError:
            return None

    @property
    def version(self) -> typing.Optional[str]:
        return self._version

    def __len__(self):
        return len(self._current_terms)


def create_minimal_ontology(graph: OntologyGraph[ID],
                            terms: typing.Sequence[MINIMAL_TERM],
                            version: typing.Optional[str] = None) -> MinimalOntology[ID, MINIMAL_TERM]:
    """
    Create minimal ontology from the components.

    :param graph: the ontology graph.
    :param terms: ALL ontology terms (both obsolete and primary).
    :param version: ontology version or `None` if unknown.
    :return: the ontology
    """
    current_terms = [term for term in terms if term.is_current]
    term_id_to_term = make_term_id_map(current_terms)
    return DefaultMinimalOntology(graph, current_terms, term_id_to_term, version)


def create_ontology(graph: OntologyGraph[ID],
                    terms: typing.Sequence[TERM],
                    version: typing.Optional[str] = None) -> Ontology[ID, TERM]:
    """
    Create ontology from the components.

    :param graph: the ontology graph.
    :param terms: ALL ontology terms (both obsolete and primary).
    :param version: ontology version or `None` if unknown.
    :return: the ontology.
    """
    current_terms = [term for term in terms if term.is_current]
    term_id_to_term = make_term_id_map(current_terms)
    return DefaultOntology(graph, current_terms, term_id_to_term, version)


def make_term_id_map(terms: typing.Sequence[MINIMAL_TERM]) -> typing.Mapping[ID, MINIMAL_TERM]:
    """
    Create a mapping from primary and alternate IDs to `MINIMAL_TERM`.

    :param terms: current ontology terms
    :return: mapping from primary and alternate IDs to `MINIMAL_TERM`
    """
    data = {}
    for term in terms:
        data[term.identifier] = term
        for alt_id in term.alt_term_ids:
            data[alt_id] = term
    return data


def _validate_term_id(term_id: CURIE_OR_TERM_ID_OR_IDENTIFIED) -> TermId:
    """
    Validate that `term_id` is a `TermId` or a valid CURIE `str`.
    """
    if isinstance(term_id, TermId):
        return term_id
    elif isinstance(term_id, Identified):
        return term_id.identifier
    elif isinstance(term_id, str):
        return TermId.from_curie(term_id)
    else:
        raise ValueError(f'Expected a `str`, a `TermId` or an `Identified` entity but got {type(term_id)}')



import functools
import logging
import re
import typing

from hpotk.constants.hpo.base import PHENOTYPIC_ABNORMALITY
from hpotk.model import TermId
from hpotk.ontology import MinimalOntology
from ._ic import AnnotationIcContainer
from ._model import SimilarityContainer
from .._traversal import get_descendants, get_ancestors, get_children

# implement Resnik IC computation starting from a collection of world documents
# and see if we can plug in excluded into the values.

HPO_PATTERN = re.compile(r"HP:(?P<ID>\d{7})")

logger = logging.getLogger('hpotk.algorithm.similarity')


def _get_common_ancestors(hpo: MinimalOntology,
                          left: TermId,
                          right: TermId) -> typing.FrozenSet[TermId]:
    la = get_ancestors(hpo, left, include_source=True)
    ra = get_ancestors(hpo, right, include_source=True)
    return la.intersection(ra)


def precalculate_ic_mica_for_hpo_concept_pairs(ic: AnnotationIcContainer,
                                               hpo: MinimalOntology) -> SimilarityContainer:
    """
    Precalculate Resnik semantic similarity for HPO :class:`TermId` pairs.

    :param ic: a mapping for obtaining an information content of a :class:`TermId`.
    :param hpo: HPO ontology.
    :return: a mapping with Resnik similarity for :class:`TermId` pairs where the similarity :math:`s>0`.
    """
    metadata = {'hpo_version': hpo.version}
    metadata.update(ic.metadata)
    data = SimilarityContainer(metadata=metadata)
    groups = list(get_children(hpo, PHENOTYPIC_ABNORMALITY))
    count = 0
    for section_top in groups:
        term = hpo.get_term(section_top)
        term_ids = list(get_descendants(hpo, section_top, include_source=True))
        logger.info(f'Calculating for {term.name} with {len(term_ids) - 1} descendants')
        for i in range(len(term_ids)):
            left = term_ids[i]
            for j in range(i, len(term_ids)):
                right = term_ids[j]

                ic_mica = functools.reduce(max,
                                           map(lambda term_id: ic.get(term_id, 0.),
                                               _get_common_ancestors(hpo, left, right)),
                                           0.)

                if left.value < right.value:
                    a, b = left.value, right.value
                else:
                    a, b = right.value, left.value

                if ic_mica > 0.:
                    previous = data.get_similarity(a, b)
                    data.set_similarity(a, b, max(ic_mica, previous))
                    count += 1
                if count % 5_000 == 0 and count != 0:
                    logger.info(f'Processed {count:,d} term pairs')

    return data



import typing

from hpotk.graph import GraphAware, OntologyGraph
from hpotk.model import TermId
from ._traversal import get_ancestors, get_descendants


def augment_with_ancestors(g: typing.Union[GraphAware, OntologyGraph],
                           source: typing.Union[TermId, typing.Collection[TermId]],
                           include_source: bool = False) -> typing.FrozenSet[TermId]:
    """
    Get a set of ancestors of the source :class:`TermId`\\ (s).
    The ancestor set may or may not include the source term IDs depending on the value of `include_source` argument.

    :param g: an ontology graph or an ontology graph-aware item
    :param source: source term ID or a collection of term IDs.
    :param include_source:  whether to include the `source` term ID(s) in the results
    :return: a :class:`frozenset` with ancestor :class:`TermId`\ s
    """
    return _augment_impl(g, source, include_source, get_ancestors)


def augment_with_descendants(g: typing.Union[GraphAware, OntologyGraph],
                             source: typing.Union[TermId, typing.Collection[TermId]],
                             include_source: bool = False) -> typing.FrozenSet[TermId]:
    """
    Get a set of descendants of the source :class:`TermId`\\ (s).
    The descendant set may or may not include the source term IDs depending on the value of `include_source` argument.

    :param g: an ontology graph or an ontology graph-aware item
    :param source: source term ID or a collection of term IDs.
    :param include_source:  whether to include the `source` term ID(s) in the results
    :return: a :class:`frozenset` with descendant :class:`TermId`\ s
    """
    return _augment_impl(g, source, include_source, get_descendants)


def _augment_impl(g: typing.Union[GraphAware, OntologyGraph],
                  source: typing.Union[TermId, typing.Collection[TermId]],
                  include_source: bool,
                  func: typing.Callable[[typing.Union[GraphAware, OntologyGraph], TermId, bool], typing.FrozenSet[TermId]]) \
        -> typing.FrozenSet[TermId]:
    if not (isinstance(g, GraphAware) or isinstance(g, OntologyGraph)):
        raise ValueError(f'hpo must be instance of GraphAware or an OntologyGraph but was {type(g)}')
    if isinstance(source, TermId):
        return get_ancestors(g, source, include_source)
    elif isinstance(source, typing.Collection):
        augmented_term_ids = set()
        for term_id in source:
            augmented_term_ids.update(func(g, term_id, include_source))
        return frozenset(augmented_term_ids)
    else:
        raise ValueError(f'source should be a TermId or a Collection of TermIds but got a {type(source)}')



from hpotk.model import TermId

# Selected descendents of Mode of inheritance `HP:0000005`.

INHERITANCE_MODIFIER: TermId = TermId.from_curie("HP:0034335")
# Inheritance modifier descendents can be added if required.

#  ########################## Mendelian inheritance and its descendents #############################################  #
MENDELIAN_INHERITANCE: TermId = TermId.from_curie("HP:0034345")

PSEUDOAUTOSOMAL_INHERITANCE: TermId = TermId.from_curie("HP:0034339")
PSEUDOAUTOSOMAL_DOMINANT_INHERITANCE: TermId = TermId.from_curie("HP:0034340")
PSEUDOAUTOSOMAL_RECESSIVE_INHERITANCE: TermId = TermId.from_curie("HP:0034341")

X_LINKED_INHERITANCE: TermId = TermId.from_curie("HP:0001417")
X_LINKED_DOMINANT_INHERITANCE: TermId = TermId.from_curie("HP:0001423")
X_LINKED_RECESSIVE_INHERITANCE: TermId = TermId.from_curie("HP:0001419")


SEMIDOMINANT_INHERITANCE: TermId = TermId.from_curie("HP:0032113")
Y_LINKED_INHERITANCE: TermId = TermId.from_curie("HP:0001450")
AUTOSOMAL_DOMINANT_INHERITANCE: TermId = TermId.from_curie("HP:0000006")
AUTOSOMAL_RECESSIVE_INHERITANCE: TermId = TermId.from_curie("HP:0000007")
MITOCHONDRIAL_INHERITANCE: TermId = TermId.from_curie("HP:0001427")
#  ##################################################################################################################  #

#  ########################## Multifactorial inheritance and its descendents ########################################  #
MULTIFACTORIAL_INHERITANCE: TermId = TermId.from_curie("HP:0001426")
OLIGOGENIC_INHERITANCE: TermId = TermId.from_curie("HP:0010983")
POLYGENIC_INHERITANCE: TermId = TermId.from_curie("HP:0010982")
DIGENIC_INHERITANCE: TermId = TermId.from_curie("HP:0010984")
#  ##################################################################################################################  #

SPORADIC: TermId = TermId.from_curie("HP:0003745")
SOMATIC_MOSAICISM: TermId = TermId.from_curie("HP:0001442")



import abc
import enum
import logging
import typing

logger = logging.getLogger("hpotk.ontology.io.obographs")


class NodeType(enum.Enum):
    CLASS = 1
    INDIVIDUAL = 2
    PROPERTY = 3


def create_property_value(cls, data: typing.Dict):
    if cls == PropertyValue:
        return PropertyValue(pred=get_attr_or_none(data, 'pred'),
                             val=get_attr_or_none(data, 'val'),
                             xrefs=data['xrefs'] if 'xrefs' in data else [],
                             meta=create_meta(data['meta']) if 'meta' in data else None)
    elif cls == DefinitionPropertyValue:
        return DefinitionPropertyValue(pred=get_attr_or_none(data, 'pred'),
                                       val=get_attr_or_none(data, 'val'),
                                       xrefs=data['xrefs'] if 'xrefs' in data else [],
                                       meta=create_meta(data['meta']) if 'meta' in data else None)
    elif cls == BasicPropertyValue:
        return BasicPropertyValue(pred=get_attr_or_none(data, 'pred'),
                                  val=get_attr_or_none(data, 'val'),
                                  xrefs=data['xrefs'] if 'xrefs' in data else [],
                                  meta=create_meta(data['meta']) if 'meta' in data else None)
    elif cls == XrefPropertyValue:
        return XrefPropertyValue(lbl=get_attr_or_none(data, 'lbl'),
                                 pred=get_attr_or_none(data, 'pred'),
                                 val=get_attr_or_none(data, 'val'),
                                 xrefs=data['xrefs'] if 'xrefs' in data else [],
                                 meta=create_meta(data['meta']) if 'meta' in data else None)
    elif cls == SynonymPropertyValue:
        return SynonymPropertyValue(synonym_type=get_attr_or_none(data, 'synonymType'),
                                    pred=get_attr_or_none(data, 'pred'),
                                    val=get_attr_or_none(data, 'val'),
                                    xrefs=data['xrefs'] if 'xrefs' in data else [],
                                    meta=create_meta(data['meta']) if 'meta' in data else None)
    else:
        return None


class PropertyValue:

    def __init__(self, pred: typing.Optional[str],
                 val: typing.Optional[str],
                 xrefs: typing.Sequence[str],
                 meta):
        self._pred = pred
        self._val = val
        self._xrefs = xrefs
        self._meta = meta

    @property
    def pred(self) -> typing.Optional[str]:
        return self._pred

    @property
    def val(self) -> typing.Optional[str]:
        return self._val

    @property
    def xrefs(self) -> typing.Sequence[str]:
        return self._xrefs

    @property
    def meta(self):
        return self._meta

    def __str__(self):
        return f'PropertyValue(pred={self.pred}, val={self.val}, xrefs={self.xrefs}, meta={self.meta})'

    def __repr__(self):
        return str(self)


class DefinitionPropertyValue(PropertyValue):

    def __str__(self):
        return f'DefinitionPropertyValue(pred={self.pred}, val={self.val}, xrefs={self.xrefs}, meta={self.meta})'

    def __repr__(self):
        return str(self)


class BasicPropertyValue(PropertyValue):

    def __str__(self):
        return f'BasicPropertyValue(pred={self.pred}, val={self.val}, xrefs={self.xrefs}, meta={self.meta})'

    def __repr__(self):
        return str(self)


class XrefPropertyValue(PropertyValue):

    def __init__(self, lbl: typing.Optional[str],
                 pred: typing.Optional[str],
                 val: typing.Optional[str],
                 xrefs: typing.Sequence[str],
                 meta):
        super().__init__(pred, val, xrefs, meta)
        self._lbl = lbl

    @property
    def lbl(self) -> typing.Optional[str]:
        return self._lbl

    def __str__(self):
        return f'XrefPropertyValue(lbl={self.lbl}, pred={self.pred}, val={self.val}, xrefs={self.xrefs}, meta={self.meta})'

    def __repr__(self):
        return str(self)


class SynonymPropertyValue(PropertyValue):

    def __init__(self, synonym_type: typing.Optional[str],
                 pred: typing.Optional[str],
                 val: typing.Optional[str],
                 xrefs: typing.Sequence[str],
                 meta):
        super().__init__(pred, val, xrefs, meta)
        self._synonym_type = synonym_type

    @property
    def synonym_type(self) -> typing.Optional[str]:
        return self._synonym_type

    def __str__(self):
        return f'SynonymPropertyValue(' \
               f'synonym_type={self.synonym_type},' \
               f' pred={self.pred},' \
               f' val={self.val},' \
               f' xrefs={self.xrefs},' \
               f' meta={self.meta})'

    def __repr__(self):
        return str(self)


class Meta:

    def __init__(self, definition: typing.Optional[DefinitionPropertyValue],
                 synonyms: typing.Sequence[SynonymPropertyValue],
                 comments: typing.Sequence[str],
                 basic_property_values: typing.Sequence[BasicPropertyValue],
                 xrefs: typing.Sequence[XrefPropertyValue],
                 is_deprecated: bool):
        self._definition = definition
        self._synonyms = synonyms
        self._comments = comments
        self._property_values = basic_property_values
        self._xrefs = xrefs
        self._is_deprecated = is_deprecated

    @property
    def definition(self) -> typing.Optional[DefinitionPropertyValue]:
        return self._definition

    @property
    def synonyms(self) -> typing.Sequence[SynonymPropertyValue]:
        return self._synonyms

    @property
    def comments(self) -> typing.Sequence[str]:
        return self._comments

    @property
    def basic_property_values(self) -> typing.Sequence[BasicPropertyValue]:
        return self._property_values

    @property
    def xrefs(self) -> typing.Sequence[XrefPropertyValue]:
        return self._xrefs

    @property
    def is_deprecated(self) -> bool:
        return self._is_deprecated

    def __str__(self):
        return f'Meta(definition={self.definition},' \
               f' synonyms={self.synonyms},' \
               f' comments={self.comments},' \
               f' basic_property_values={self.basic_property_values},' \
               f' xrefs={self.xrefs},' \
               f' is_deprecated={self.is_deprecated})'

    def __repr__(self):
        return str(self)


class NodeOrEdge(metaclass=abc.ABCMeta):

    @property
    @abc.abstractmethod
    def meta(self) -> typing.Optional[Meta]:
        pass


class Node(NodeOrEdge):

    def __init__(self, id: str,
                 lbl: typing.Optional[str],
                 node_type: NodeType,
                 meta: typing.Optional[Meta]):
        self._id = id
        self._lbl = lbl
        self._type = node_type
        self._meta = meta

    @property
    def id(self) -> str:
        return self._id

    @property
    def lbl(self) -> typing.Optional[str]:
        return self._lbl

    @property
    def type(self) -> NodeType:
        return self._type

    @property
    def meta(self) -> typing.Optional[Meta]:
        return self._meta

    def __str__(self):
        return f'Node(id={self.id}, lbl={self.lbl}, type={self.type} meta={self.meta})'

    def __repr__(self):
        return str(self)


class Edge(NodeOrEdge):

    def __init__(self, sub: str,
                 pred: str,
                 obj: str,
                 meta: typing.Optional[Meta]):
        self._sub = sub
        self._pred = pred
        self._obj = obj
        self._meta = meta

    @property
    def sub(self) -> str:
        return self._sub

    @property
    def pred(self) -> str:
        return self._pred

    @property
    def obj(self) -> str:
        return self._obj

    @property
    def meta(self) -> typing.Optional[Meta]:
        return self._meta

    def __str__(self):
        return f'Edge(sub={self.sub}, pred={self.pred}, obj={self.obj}, meta={self.meta})'

    def __repr__(self):
        return str(self)


def get_attr_or_none(data: dict, key: str):
    return data[key] if key in data else None


def create_meta(data) -> typing.Optional[Meta]:
    definition = create_property_value(DefinitionPropertyValue, data['definition']) if 'definition' in data else None
    comments = data['comments'] if 'comments' in data else []
    basic_property_values = [create_property_value(BasicPropertyValue, d) for d in
                             data['basicPropertyValues']] if 'basicPropertyValues' in data else []
    synonyms = [create_property_value(SynonymPropertyValue, x) for x in data['synonyms']] if 'synonyms' in data else []
    xrefs = [create_property_value(XrefPropertyValue, x) for x in data['xrefs']] if 'xrefs' in data else []
    is_deprecated = 'deprecated' in data

    return Meta(definition, synonyms, comments, basic_property_values, xrefs, is_deprecated)


def create_node(data) -> typing.Optional[Node]:
    identifier = data['id']
    lbl = data['lbl'] if 'lbl' in data else None

    if 'type' not in data:
        logger.debug(f'Missing node type in {data}')
        # We cannot create a Node with no type.
        return None

    nt = data['type']
    try:
        node_type = NodeType[nt]
    except KeyError:
        logger.debug(f'Unknown node type {nt}. Supported: {list(NodeType)}')
        return None

    if 'meta' in data:
        meta = create_meta(data['meta'])
    else:
        meta = None

    return Node(identifier, lbl, node_type, meta)


def create_edge(data) -> Edge:
    if not all([key in data for key in ['sub', 'pred', 'obj']]):
        raise ValueError(f'Missing required property for an edge {data}')
    meta = create_meta(data['meta']) if 'meta' in data else None
    return Edge(data['sub'], data['pred'], data['obj'], meta)



import typing

from hpotk.model import Identified, TermId, CURIE_OR_TERM_ID


class HpoFrequency(Identified):

    def __init__(self, identifier: TermId,
                 lower_bound: float,
                 upper_bound: float):
        self._id = identifier
        self._lower = lower_bound
        self._upper = upper_bound

    @property
    def identifier(self) -> TermId:
        return self._id

    @property
    def lower_bound(self) -> float:
        return self._lower

    @property
    def upper_bound(self) -> float:
        return self._upper

    @property
    def frequency(self) -> float:
        return self._lower + self._upper / 2

    def __eq__(self, other):
        return isinstance(other, HpoFrequency) \
            and self._id == other._id \
            and self._lower == other._lower \
            and self._upper == other._upper

    def __str__(self):
        return f'HpoFrequency(identifier={self.identifier.value}, ' \
               f'lower_bound={self.lower_bound}, ' \
               f'upper_bound={self.upper_bound})'

    def __repr__(self):
        return str(self)


# Children of Frequency `HP:0040279`.
FREQUENCY: TermId = TermId.from_curie("HP:0040279")

EXCLUDED: TermId = TermId.from_curie("HP:0040285")
VERY_RARE: TermId = TermId.from_curie("HP:0040284")
OCCASIONAL: TermId = TermId.from_curie("HP:0040283")
FREQUENT: TermId = TermId.from_curie("HP:0040282")
VERY_FREQUENT: TermId = TermId.from_curie("HP:0040281")
OBLIGATE: TermId = TermId.from_curie("HP:0040280")


_frequencies = (
    HpoFrequency(EXCLUDED, 0., 0.),
    HpoFrequency(VERY_RARE, .01, .04),
    HpoFrequency(OCCASIONAL, .05, .29),
    HpoFrequency(FREQUENT, .3, .79),
    HpoFrequency(VERY_FREQUENT, .8, .99),
    HpoFrequency(OBLIGATE, 1., 1.)
)

HPO_FREQUENCIES = {fq.identifier: fq for fq in _frequencies}


def parse_hpo_frequency(value: CURIE_OR_TERM_ID) -> typing.Optional[HpoFrequency]:
    """
    Get :class:`HpoFrequency` for given HPO frequency term ID or `None` if the term ID does not represent
    a frequency term.

    :param value: a `str` or :class:`TermId` corresponding to HPO frequency term ID.
    :return: :class:`HpoFrequency` or `None`.
    """
    if isinstance(value, TermId):
        pass
    elif isinstance(value, str):
        value = TermId.from_curie(value)
    else:
        return None

    try:
        return HPO_FREQUENCIES[value]
    except KeyError:
        return None



import abc
import typing

from hpotk.model import TermId, Identified


class TermIdSorting(metaclass=abc.ABCMeta):
    """
    `TermIdSorting` computes indices for sorting a sequence of identifiers/identified items.
    """

    @abc.abstractmethod
    def argsort(self, term_ids: typing.Sequence[typing.Union[TermId, Identified]]) -> typing.Sequence[int]:
        """
        Prepare indices for sorting a sequence of term IDs.

        :param term_ids: a sequence of term IDs or identified entities to sort.
        :return: a sequence of indices for sorting of the `term_ids` sequence.
        """
        pass



import re
import enum
import typing
from collections import defaultdict, namedtuple

from hpotk.annotations import HpoDiseases, EvidenceCode, AnnotationReference, Sex
from hpotk.annotations import SimpleHpoDiseaseAnnotation, SimpleHpoDisease, SimpleHpoDiseases
from hpotk.model import TermId
from hpotk.ontology import MinimalOntology
from hpotk.util import open_text_io_handle_for_reading
from hpotk.constants.hpo.frequency import parse_hpo_frequency
from hpotk.annotations.load._api import HpoDiseaseLoader

HpoAnnotationLine = namedtuple('HpoAnnotationLine',
                               field_names=[
                                   'disease_id', 'disease_name', 'is_negated',
                                   'phenotype_term_id',
                                   'annotation_references', 'onset', 'frequency',
                                   'sex', 'modifiers', 'aspect', 'curators']
                               )

HPOA_VERSION_PATTERN = re.compile(r'^#(date|version): (?P<version>[\w-]+)\w?$')
HPO_PATTERN = re.compile(r'^HP:\d{7}$')
RATIO_PATTERN = re.compile(r'^(?P<numerator>\d+)/(?P<denominator>\d+)$')
PERCENTAGE_PATTERN = re.compile(r'^(?P<value>\d+\.?(\d+)?)%$')


class Ratio:
    """
    A private helper class for parsing frequency data.
    """

    def __init__(self, numerator: int, denominator: int):
        self._numerator = numerator
        self._denominator = denominator

    @property
    def numerator(self) -> int:
        return self._numerator

    @property
    def denominator(self) -> int:
        return self._denominator

    @property
    def frequency(self) -> float:
        return self.numerator / self.denominator

    def is_positive(self) -> bool:
        return self.numerator > 0

    def is_zero(self) -> bool:
        return self.numerator == 0

    @staticmethod
    def fold(left, right):
        """
        Fold two :class:`Ratio`s together into a new :class:`Ratio` that represents `n` over `m` of both inputs.

        Note that this is *NOT* the addition of two :class:`Ratio`s!

        For instance, if :math:`n_1` of :math:`m_1` and :math:`n_2` of :math:`m_2` population members like lasagna,
        then :math:`n_1 + n_2` of :math:`m_1 + m_2` people like lasagna in total.

        :param left: left :class:`Ratio`.
        :param right: right :class:`Ratio`.
        :return: the result as a :class:`SimpleRatio`.
        """
        if isinstance(left, Ratio) and isinstance(right, Ratio):
            return Ratio(left.numerator + right.numerator, left.denominator + right.denominator)
        else:
            msg = 'left arg must be an instance of `Ratio`' \
                if isinstance(right, Ratio) \
                else 'right arg must be an instance of `Ratio`'
            raise ValueError(msg)

    def __eq__(self, other):
        return isinstance(other, Ratio) \
            and self.numerator * other.denominator == other.numerator * self.denominator

    def __str__(self):
        return f"{self.numerator}/{self.denominator}"

    def __repr__(self):
        return f"SimpleRatio(" \
               f"numerator={self._numerator}, " \
               f"denominator={self._denominator})"


class SimpleHpoaDiseaseLoader(HpoDiseaseLoader):
    """
    Loads HPO annotation file into :class:`HpoDiseases`.
    """

    def __init__(self, hpo: MinimalOntology,
                 cohort_size: int = 50,
                 salvage_negated_frequencies: bool = False):
        if not isinstance(hpo, MinimalOntology):
            raise ValueError(f'hpo must be an instance of `MinimalOntology` but was {type(hpo)}')
        self._hpo = hpo
        self._cohort_size = cohort_size
        self._salvage_negated_frequencies = salvage_negated_frequencies

    def load(self, file: typing.Union[typing.IO, str]) -> HpoDiseases:
        data: typing.Mapping[str, typing.List[HpoAnnotationLine]] = defaultdict(list)
        version = None
        expecting_to_see_header_line = True
        with open_text_io_handle_for_reading(file) as fh:
            for line in fh:
                if expecting_to_see_header_line:
                    if line.startswith('#'):
                        # header
                        if line.startswith('#DatabaseID'):
                            # The older HPOA format
                            expecting_to_see_header_line = False
                        else:
                            version_matcher = HPOA_VERSION_PATTERN.match(line)
                            if version_matcher:
                                version = version_matcher.group('version')
                    else:
                        if line.startswith('database_id'):
                            expecting_to_see_header_line = False
                    continue
                else:
                    # corpus
                    hpoa = _parse_hpoa_line(line)
                    if hpoa:
                        data[hpoa.disease_id].append(hpoa)

        diseases = []
        for disease_id, hpoa_lines in data.items():
            disease = self._assemble_hpo_disease(disease_id, hpoa_lines)
            diseases.append(disease)

        return SimpleHpoDiseases(diseases, version)

    @property
    def cohort_size(self) -> int:
        return self._cohort_size

    def _assemble_hpo_disease(self, disease_curie: str, hpoa_lines: typing.List[HpoAnnotationLine]):
        # If the hpoa_lines is empty, then there is something wrong with the `defaultdict` and the logic above.
        disease_id = TermId.from_curie(disease_curie)
        disease_name = hpoa_lines[0].disease_name
        annotations, moi = self._parse_hpo_annotations(hpoa_lines)
        return SimpleHpoDisease(disease_id, disease_name, annotations, moi)

    def _parse_hpo_annotations(self, hpoa_lines: typing.Iterable[HpoAnnotationLine]) \
            -> typing.Tuple[typing.List[SimpleHpoDiseaseAnnotation], typing.Collection[TermId]]:

        line_by_phenotype: typing.Mapping[str, typing.List[HpoAnnotationLine]] = defaultdict(list)
        moi = set()
        for hpoa in hpoa_lines:
            if hpoa.aspect == Aspect.PHENOTYPE:
                # Several HPOA lines may correspond to a single phenotype feature
                line_by_phenotype[hpoa.phenotype_term_id].append(hpoa)
            elif hpoa.aspect == Aspect.INHERITANCE:
                moi.add(hpoa.phenotype_term_id)
            else:
                # TODO - handle the remaining aspect lines
                pass

        annotations = []
        for phenotype_curie, lines in line_by_phenotype.items():
            phenotype_id = TermId.from_curie(phenotype_curie)
            total_ratio = None
            annotation_references = set()
            modifiers = set()
            for line in lines:
                ratio = self._parse_frequency(line.is_negated, line.frequency)
                if total_ratio:
                    total_ratio = Ratio.fold(total_ratio, ratio)
                else:
                    total_ratio = ratio

                annotation_references.update(line.annotation_references)
                modifiers.update(line.modifiers)

            ann = SimpleHpoDiseaseAnnotation(phenotype_id,
                                             numerator=total_ratio.numerator,
                                             denominator=total_ratio.denominator,
                                             references=tuple(annotation_references),
                                             modifiers=tuple(modifiers))
            annotations.append(ann)

        return annotations, moi

    def _parse_frequency(self, is_negated: bool, frequency: str) -> Ratio:
        # An empty string is assumed to represent a case study
        if not frequency:
            numerator = 0 if is_negated else 1
            denominator = 1
            return Ratio(numerator, denominator)

        # HPO term, e.g. HP:0040280 (Obligate)
        hpo_match = HPO_PATTERN.match(frequency)
        if hpo_match:
            hpo_frequency = parse_hpo_frequency(frequency)
            numerator = 0 if is_negated else round(hpo_frequency.frequency * self._cohort_size)
            denominator = self._cohort_size
            return Ratio(numerator, denominator)

        # Ratio, e.g. 1/2
        ratio_match = RATIO_PATTERN.match(frequency)
        if ratio_match:
            denominator = int(ratio_match.group('denominator'))
            i = int(ratio_match.group('numerator'))
            if is_negated:
                if denominator == 0:
                    # fix denominator in cases like 0/0
                    denominator = self._cohort_size
                if i == 0 and self._salvage_negated_frequencies:
                    numerator = 0
                else:
                    numerator = denominator - i
            else:
                numerator = i

            return Ratio(numerator, denominator)

        # Percentage, e.g. 20%
        percentage_match = PERCENTAGE_PATTERN.match(frequency)
        if percentage_match:
            percentage = float(percentage_match.group('value'))
            numerator = round(percentage * self._cohort_size / 100)
            denominator = self._cohort_size
            return Ratio(numerator, denominator)

        raise ValueError(f'Unable to parse frequency {frequency}')


def _parse_hpoa_line(line: str) -> typing.Optional[HpoAnnotationLine]:
    fields = line.strip().split('\t')

    disease_id = fields[0]
    disease_name = fields[1]
    is_negated = fields[2].upper() == 'NOT'
    phenotype_id = fields[3]
    evidence_code = EvidenceCode.parse(fields[5])
    annotation_references = [AnnotationReference(TermId.from_curie(term_id), evidence_code)
                             for term_id
                             in filter(lambda t: t and not t.isspace(), fields[4].split(';'))]
    # TODO - implement parsing of temporal data
    onset = None

    frequency = fields[7]
    sex = Sex.parse(fields[8])

    modifiers = [TermId.from_curie(term_id)
                 for term_id
                 in filter(lambda t: t and not t.isspace(), fields[9].split(';'))]
    aspect = Aspect.parse(fields[10])
    curators = [curator.strip() for curator in fields[11].split(';')]

    return HpoAnnotationLine(disease_id, disease_name, is_negated,
                             phenotype_id,
                             annotation_references, onset, frequency,
                             sex, modifiers, aspect, curators)


class Aspect(enum.Enum):
    """Phenotype."""
    PHENOTYPE = 0
    """Inheritance."""
    INHERITANCE = 1
    """C"""
    C = 2
    """Modifier."""
    MODIFIER = 3

    @staticmethod
    def parse(value: str):
        """
        Parse :class:`Aspect` from `str` value.

        :param value: a `str` with the aspect code.
        :return: the parsed enum member or `None` if `value` is not valid :class:`Aspect` value.
        """
        value = value.upper()
        if value == 'P':
            return Aspect.PHENOTYPE
        elif value == 'I':
            return Aspect.INHERITANCE
        elif value == 'C':
            return Aspect.C
        elif value == 'M':
            return Aspect.MODIFIER
        else:
            return None



import typing

import numpy as np

from hpotk.model import TermId

from ._api import OntologyGraph

from ._csr_graph import BisectPoweredCsrOntologyGraph
from .csr import ImmutableCsrMatrix


def get_toy_graph() -> typing.Tuple[typing.Sequence[TermId], OntologyGraph]:
    root = TermId.from_curie('HP:1')
    curies = [
        'HP:01', 'HP:010', 'HP:011', 'HP:0110',
        'HP:02', 'HP:020', 'HP:021', 'HP:022',
        'HP:03', 'HP:1'
    ]
    nodes = np.fromiter(map(TermId.from_curie, curies), dtype=object)
    row = [0, 3, 5, 7, 9, 13, 14, 15, 16, 17, 20]
    col = [1, 2, 9, 0, 3, 0, 3, 1, 2, 5, 6, 7, 9, 4, 4, 4, 9, 0, 4, 8]
    data = [-1, -1, 1, 1, -1, 1, -1, 1, 1, -1, -1, -1, 1, 1, 1, 1, 1, -1, -1, -1]
    am = ImmutableCsrMatrix(row, col, data, shape=(len(nodes), len(nodes)), dtype=int)

    return nodes, BisectPoweredCsrOntologyGraph(root, nodes, am)


import typing


T = typing.TypeVar('T')


def validate_instance(obj: T,
                      clz: type,
                      param_name: typing.Optional[str] = None) -> T:
    """
    Validate that `obj` is instance of `clz` or raise :class:`ValueError` otherwise.

    :param obj: and instance for validation.
    :param clz: the target class.
    :param param_name: name of the object to include in the error message.
    :return: the `obj` if the validation passes.
    :raise ValueError: if the instance check fails.
    """
    if not isinstance(obj, clz):
        if param_name is None:
            raise ValueError(f'The object must be an instance of {clz} but was {type(obj)}')
        else:
            raise ValueError(f'{param_name} must be an instance of {clz} but was {type(obj)}')
    return obj


def validate_optional_instance(obj: typing.Optional[T],
                               clz: type,
                               param_name: typing.Optional[str] = None) -> typing.Optional[T]:
    """
    Validate that `obj` is instance of `clz` or `None`, or raise :class:`ValueError` otherwise.

    :param obj: and instance for validation or `None`.
    :param clz: the target class.
    :param param_name: name of the object to include in the error message.
    :return: the `obj` if the validation passes.
    :raise ValueError: if `obj` is *not* `None` and the instance check fails.
    """
    if obj is not None:
        return validate_instance(obj, clz, param_name)



import json
import os
import typing
import urllib
from urllib.request import urlopen, HTTPError

from hpotk.ontology import Ontology, MinimalOntology
from hpotk.ontology.load.obographs import load_minimal_ontology, load_ontology
from ._api import OntologyType, OntologyStore


class GitHubOntologyStore(OntologyStore):
    """
    `GitHubOntologyStore` fetches an Obographs ontology JSON from GitHub.
    """

    ONTOLOGY_CREDENTIALS = {
        OntologyType.HPO: {
            'owner': 'obophenotype',
            'repo': 'human-phenotype-ontology',
        }
    }

    def __init__(
            self,
            store_dir: str,
            timeout: int = 10,
    ):
        super().__init__(store_dir=store_dir)
        self._timeout = timeout
        self._tag_api_url = 'https://api.github.com/repos/{owner}/{repo}/tags'
        self._release_url = 'https://github.com/{owner}/{repo}/releases/download/{release}/{ontology_id}.json'

    def load_minimal_ontology(self, ontology_type: OntologyType,
                              release: typing.Optional[str] = None) -> MinimalOntology:
        return self._impl_load_ontology(
            load_minimal_ontology,
            ontology_type,
            release,
        )

    def load_ontology(
            self,
            ontology_type: OntologyType,
            release: typing.Optional[str] = None,
    ) -> Ontology:
        return self._impl_load_ontology(
            load_ontology,
            ontology_type,
            release,
        )

    def _impl_load_ontology(
            self,
            loader_func,
            ontology_type: OntologyType,
            release: typing.Optional[str] = None,
    ):
        if ontology_type not in self.ONTOLOGY_CREDENTIALS:
            raise ValueError(f'Ontology {ontology_type} not among the known ontology credentials')
        credentials = self.ONTOLOGY_CREDENTIALS[ontology_type]

        # Figure out the desired release
        if release is None:
            release = self._fetch_latest_tag_from_github(credentials)
        self._logger.debug('Using %s as the ontology release', release)

        # Check if we have the release in the local storage
        # and download the JSON file if not.
        fpath_ontology = self._download_ontology_if_missing(credentials, ontology_type.identifier, release)

        # Load the ontology
        return loader_func(fpath_ontology)

    def _download_ontology_if_missing(
            self,
            credentials: typing.Mapping[str, str],
            ontology_id: str,
            release: str,
    ) -> str:
        fdir_ontology = os.path.join(self.store_dir, ontology_id)
        fpath_ontology = os.path.join(fdir_ontology, f'{ontology_id.lower()}.{release}.json')
        if not os.path.isfile(fpath_ontology):
            os.makedirs(fdir_ontology, exist_ok=True)
            owner = credentials['owner']
            repo = credentials['repo']
            url = self._release_url.format(
                owner=owner,
                repo=repo,
                release=release,
                ontology_id=ontology_id.lower(),
            )
            self._logger.info('Downloading ontology from %s', url)
            self._logger.info('Storing the ontology at %s', fpath_ontology)
            try:
                with urlopen(url, timeout=self._timeout) as response, open(fpath_ontology, 'wb') as fh_ontology:
                    fh_ontology.write(response.read())
            except HTTPError as he:
                if he.code == 404:
                    # Most likely a non-existing release.
                    raise ValueError(f'Could not find {release} on GitHub')
                else:
                    # Another error.
                    raise he
            self._logger.info('Download complete')

        return fpath_ontology

    def _fetch_latest_tag_from_github(self, credentials: typing.Mapping[str, str]):
        self._logger.debug('Release unset, getting the latest')
        tag_names = self._get_tag_names(
            owner=credentials['owner'],
            repo=credentials['repo'],
        )
        # We assume lexicographic sorting of the tags
        return max(tag_names)

    def _get_tag_names(self, owner: str, repo: str) -> typing.Iterable[str]:
        tag_url = self._tag_api_url.format(owner=owner, repo=repo)
        self._logger.debug('Pulling tag from %s', tag_url)

        with urllib.request.urlopen(tag_url, timeout=self._timeout) as r:
            tags = json.load(r)

        if len(tags) == 0:
            raise ValueError('No tags could be fetched from GitHub tag API')
        else:
            self._logger.debug('Fetched %d tags', len(tags))

        return [tag['name'] for tag in tags]



import typing

import pytest

import hpotk

from hpotk.model import Identified, ObservableFeature, TermId
from hpotk.validate import ValidationResult, ValidationLevel
from hpotk.validate import AnnotationPropagationValidator, PhenotypicAbnormalityValidator, ObsoleteTermIdsValidator



class SimpleFeature(Identified, ObservableFeature):

    def __init__(self, curie: str, status: bool):
        self._id = TermId.from_curie(curie)
        self._status = status

    @property
    def identifier(self) -> TermId:
        return self._id

    @property
    def is_present(self) -> bool:
        return self._status


@pytest.fixture
def example_terms() -> typing.Sequence[SimpleFeature]:
    return (
        SimpleFeature('HP:0001166', True),  # Arachnodactyly
        SimpleFeature('HP:0001250', True),  # Seizure
        SimpleFeature('HP:0032648', True),  # Tubularization of Bowman capsule
        SimpleFeature('HP:0000805', False)  # excluded Enuresis
    )


class TestAnnotationPropagationValidator:

    @pytest.fixture
    def validator(self, toy_hpo: hpotk.MinimalOntology) -> AnnotationPropagationValidator:
        return AnnotationPropagationValidator(toy_hpo)

    def test_empty_input_is_allowed(self, validator: AnnotationPropagationValidator):
        results = validator.validate([])
        assert results.is_ok()

    def test_ok_input_produces_no_errors(self, validator: AnnotationPropagationValidator,
                                         example_terms: typing.Sequence[SimpleFeature]):
        results = validator.validate(example_terms)
        assert results.is_ok()

    def test_obsolete_ancestor_produces_error(self, validator: AnnotationPropagationValidator,
                                              example_terms: typing.Sequence[SimpleFeature]):
        example_terms = list(example_terms)
        example_terms.append(TermId.from_curie('HP:0006010'))

        results = validator.validate(example_terms)

        assert not results.is_ok()
        assert results.results[0] == ValidationResult(level=ValidationLevel.ERROR, category='annotation_propagation',
                                                      message='Terms should not contain both present Arachnodactyly [HP:0001166] '
                                                              f'and its present or excluded ancestor Long fingers [HP:0100807]')

    @pytest.mark.parametrize('ancestor_curie, ancestor_status, base_curie, base_status',
                             (
                                     # present Long fingers vs. present Arachnodactyly
                                     ("HP:0100807", True, 'HP:0001166', True),

                                     # excluded Long fingers vs. present Arachnodactyly
                                     ("HP:0100807", False, 'HP:0001166', True),

                                     # The case is commented out since it is possible to have some abnormality but excluded specific abnormality
                                     # ("HP:0000014", True, 'HP:0000805', False),  # present Abnormality of the bladder vs. excluded Enuresis

                                     # excluded Abnormality of the bladder vs. excluded Enuresis
                                     ("HP:0000014", False, 'HP:0000805', False),
                             ))
    def test_ancestor_presence_produces_error(self, ancestor_curie, ancestor_status, base_curie, base_status,
                                              toy_hpo: hpotk.MinimalOntology, validator: AnnotationPropagationValidator,
                                              example_terms: typing.Sequence[SimpleFeature]):
        example_terms = list(example_terms)
        example_terms.append(SimpleFeature(ancestor_curie, ancestor_status))

        results = validator.validate(example_terms)

        assert not results.is_ok()
        assert 1 == len(results.results)
        first = results.results[0]
        assert ValidationLevel.ERROR == first.level
        assert 'annotation_propagation' == first.category

        state = 'present' if base_status else 'excluded'
        assert (f'Terms should not contain both {state} '
                f'{toy_hpo.get_term(base_curie).name} [{toy_hpo.get_term(base_curie).identifier.value}] '
                f'and its present or excluded ancestor '
                f'{toy_hpo.get_term(ancestor_curie).name} [{ancestor_curie}]' == first.message)


class TestPhenotypicAbnormalityValidator:

    @pytest.fixture
    def validator(self, toy_hpo: hpotk.MinimalOntology) -> PhenotypicAbnormalityValidator:
        return PhenotypicAbnormalityValidator(toy_hpo)

    def test_ok_input_produces_no_errors(self, validator: PhenotypicAbnormalityValidator,
                                         example_terms: typing.Sequence[SimpleFeature]):
        results = validator.validate(example_terms)
        assert results.is_ok()

    @pytest.mark.parametrize('curie',
                             (
                                     "HP:0012823",  # Clinical modifier
                                     "HP:0003825",  # Variable expressivity
                                     "HP:0003621",  # Juvenile onset
                             )
                             )
    def test_presence_of_clinical_modifier_produces_error(self, curie,
                                                          validator: PhenotypicAbnormalityValidator,
                                                          toy_hpo: hpotk.MinimalOntology,
                                                          example_terms: typing.Sequence[SimpleFeature]):
        example_terms = list(example_terms)
        example_terms.append(SimpleFeature(curie, True))
        results = validator.validate(example_terms)

        assert not results.is_ok()
        assert results.results[0] == ValidationResult(level=ValidationLevel.WARNING,
                                                      category='phenotypic_abnormality_descendant',
                                                      message=f'{toy_hpo.get_term(curie).name} '
                                                              f'[{toy_hpo.get_term(curie).identifier.value}] '
                                                              f'is not a descendant of '
                                                              'Phenotypic abnormality [HP:0000118]')


class TestObsoleteTermIdsValidator:

    @pytest.fixture
    def validator(self, toy_hpo: hpotk.MinimalOntology) -> ObsoleteTermIdsValidator:
        return ObsoleteTermIdsValidator(toy_hpo)

    def test_ok_input_produces_no_errors(self, validator: ObsoleteTermIdsValidator,
                                         example_terms: typing.Sequence[SimpleFeature]):
        results = validator.validate(example_terms)

        assert results.is_ok()

    def test_presence_of_an_obsolete_term_produces_error(self, validator: ObsoleteTermIdsValidator,
                                                         toy_hpo: hpotk.MinimalOntology,
                                                         example_terms: typing.Sequence[SimpleFeature]):
        curie = 'HP:0006010'  # obsolete Long fingers
        example_terms = list(example_terms)
        example_terms.append(SimpleFeature(curie, True))
        results = validator.validate(example_terms)

        assert not results.is_ok()
        assert results.results[0] == ValidationResult(level=ValidationLevel.WARNING,
                                                      category='obsolete_term_id_is_used',
                                                      message=f'Using the obsolete {curie} instead of '
                                                              f'{toy_hpo.get_term(curie).identifier.value} '
                                                              f'for {toy_hpo.get_term(curie).name}')



from hpotk.model import TermId

# Children of All `HP:0000001`, the root of the HPO hierarchy.

PHENOTYPIC_ABNORMALITY: TermId = TermId.from_curie('HP:0000118')
"""
Phenotypic abnormality is the parent term of all phenotypic features of HPO.
"""

CLINICAL_MODIFIER: TermId = TermId.from_curie("HP:0012833")
"""
Clinical modifier is the parent term of the clinical modifier HPO module.
"""

MODE_OF_INHERITANCE: TermId = TermId.from_curie("HP:0000005")
"""
Mode of inheritance is the parent term of the inheritance mode subgraph.
"""

PAST_MEDICAL_HISTORY: TermId = TermId.from_curie("HP:0032443")
BLOOD_GROUP: TermId = TermId.from_curie("HP:0032223")
FREQUENCY: TermId = TermId.from_curie("HP:0040279")



import abc
import enum
import logging
import typing

from hpotk.ontology import MinimalOntology, Ontology


class OntologyType(enum.Enum):
    """
    Enum with the ontologies supported by the :class:`OntologyStore`.
    """

    HPO = 'HPO', 'HP'
    """
    Human Phenotype Ontology.
    """

    def __new__(cls, *args, **kwargs):
        obj = object.__new__(cls)
        obj._value_ = args[0]
        return obj

    def __init__(self, _: str, identifier: str):
        self._id_ = identifier

    @property
    def identifier(self) -> str:
        """
        Get a `str` with the ontology identifier (e.g. ``HP`` for HPO).
        """
        return self._id_


class OntologyStore(metaclass=abc.ABCMeta):
    """
    `OntologyStore` stores versions of the supported ontologies.
    """

    def __init__(
            self,
            store_dir: str,
    ):
        self._logger = logging.getLogger(__name__)
        self._store_dir = store_dir

    @abc.abstractmethod
    def load_minimal_ontology(
            self,
            ontology_type: OntologyType,
            release: typing.Optional[str] = None,
    ) -> MinimalOntology:
        """
        Load a `release` of a given `ontology_type` as a minimal ontology.

        :param ontology_type: the desired ontology type, see :class:`OntologyType` for a list of supported ontologies.
        :param release: a `str` with the ontology release tag or `None` if the latest ontology should be fetched.
        :return: a minimal ontology.
        """
        pass

    @abc.abstractmethod
    def load_ontology(
            self,
            ontology_type: OntologyType,
            release: typing.Optional[str] = None,
    ) -> Ontology:
        """
        Load a `release` of a given `ontology_type` as an ontology.

        :param ontology_type: the desired ontology type, see :class:`OntologyType` for a list of supported ontologies.
        :param release: a `str` with the ontology release tag or `None` if the latest ontology should be fetched.
        :return: an ontology.
        """
        pass

    @property
    def store_dir(self) -> str:
        """
        Get a `str` with a platform specific absolute path to the data directory.

        The data directory points to `$HOME/.hpo-toolkit` on UNIX and `$HOME/hpo-toolkit` on Windows.
        The folder is created if it does not exist.
        """
        return self._store_dir

    def load_minimal_hpo(
            self,
            release: typing.Optional[str] = None,
    ) -> MinimalOntology:
        """
        A convenience method for loading a specific HPO release.

        :param release: an optional `str` with the desired HPO release (if `None`, the latest HPO will be provided).
        :return: a :class:`hpotk.MinimalOntology` with the HPO data.
        """
        return self.load_minimal_ontology(OntologyType.HPO, release=release)

    def load_hpo(
            self,
            release: typing.Optional[str] = None,
    ) -> Ontology:
        """
        A convenience method for loading a specific HPO release.

        :param release: an optional `str` with the desired HPO release (if `None`, the latest HPO will be provided).
        :return: a :class:`hpotk.Ontology` with the HPO data.
        """
        return self.load_ontology(OntologyType.HPO, release=release)



import typing
import warnings

from hpotk.model import TermId, CURIE_OR_TERM_ID
from ._api import ANNOTATED_ITEM
from ._base import AnnotationReference
from ._base import HpoDisease, HpoDiseaseAnnotation, HpoDiseases


class SimpleHpoDiseaseAnnotation(HpoDiseaseAnnotation):

    def __init__(self, identifier: TermId,
                 numerator: int,
                 denominator: int,
                 references: typing.Sequence[AnnotationReference],
                 modifiers: typing.Sequence[TermId]):
        self._id = identifier
        self.check_numerator_and_denominator(numerator, denominator)
        self._numerator = numerator
        self._denominator = denominator
        self._refs = references
        self._modifiers = modifiers

    @property
    def identifier(self) -> TermId:
        return self._id

    @property
    def numerator(self) -> int:
        return self._numerator

    @property
    def denominator(self) -> int:
        return self._denominator

    @property
    def references(self) -> typing.Sequence[AnnotationReference]:
        return self._refs

    @property
    def modifiers(self) -> typing.Sequence[TermId]:
        return self._modifiers

    def __repr__(self):
        return f"SimpleHpoDiseaseAnnotation(" \
               f"identifier={self.identifier}, " \
               f"numerator={self.numerator}, " \
               f"denominator={self.denominator}, " \
               f"references={self.references}, " \
               f"modifiers={self.modifiers})"


class SimpleHpoDisease(HpoDisease):

    def __init__(self, identifier: TermId,
                 name: str,
                 annotations: typing.Collection[HpoDiseaseAnnotation],
                 modes_of_inheritance: typing.Collection[TermId]):
        self._id = identifier
        self._name = name
        self._annotations = annotations
        self._modes_of_inheritance = modes_of_inheritance

    @property
    def identifier(self) -> TermId:
        return self._id

    @property
    def name(self) -> str:
        return self._name

    @property
    def annotations(self) -> typing.Collection[HpoDiseaseAnnotation]:
        return self._annotations

    @property
    def modes_of_inheritance(self) -> typing.Collection[TermId]:
        return self._modes_of_inheritance


class SimpleHpoDiseases(HpoDiseases):

    def __init__(self, diseases: typing.Iterable[HpoDisease], version: str = None):
        self._diseases = {d.identifier: d for d in diseases}
        self._version = version

    @property
    def items(self) -> typing.Collection[ANNOTATED_ITEM]:
        return self._diseases.values()

    @property
    def diseases(self) -> typing.Collection[HpoDisease]:
        # REMOVE(v1.0.0)
        warnings.warn('The `diseases` property has been deprecated and will be removed in `v1.0.0`. '
                      'Use `items()` instead',
                      category=DeprecationWarning, stacklevel=2)
        return self._diseases.values()

    @property
    def disease_ids(self):
        # REMOVE(v1.0.0)
        warnings.warn(f'`disease_ids` property has been deprecated and will be removed in v1.0.0. '
                      f'Iterate over `item_ids()` instead.',
                      DeprecationWarning, stacklevel=2)
        return list(self.item_ids())

    def __getitem__(self, item: CURIE_OR_TERM_ID) -> typing.Optional[HpoDisease]:
        if isinstance(item, TermId):
            pass
        elif isinstance(item, str):
            item = TermId.from_curie(item)
        else:
            raise ValueError(f'Expected a `str` or `TermId` but got {type(item)}')
        try:
            return self._diseases[item]
        except KeyError:
            return None

    def __iter__(self) -> typing.Iterator[HpoDiseaseAnnotation]:
        return iter(self.items)

    def __len__(self) -> int:
        return len(self._diseases)

    @property
    def version(self) -> typing.Optional[str]:
        return self._version



"""
The `hpotk.ontology` package defines what an ontology is and provides methods for working with ontology data.
"""

from ._api import MinimalOntology, Ontology
from ._default import create_minimal_ontology, create_ontology
from . import load

__all__ = [
    'MinimalOntology', 'Ontology',
    'create_minimal_ontology', 'create_ontology'
]



import pytest

from hpotk.model import *


class TestTermId:

    @pytest.mark.parametrize('curie',
                             ("HP:1234567", "HP_1234567")
                             )
    def test_from_curie(self, curie):
        term_id = TermId.from_curie(curie)
        assert term_id.value == "HP:1234567"

    def test_from_curie__unusual_input(self):
        term_id = TermId.from_curie('SNOMEDCT_US:313307000')
        assert term_id.prefix == 'SNOMEDCT_US'
        assert term_id.id == '313307000'

    @pytest.mark.parametrize('curie, message',
                             (("HP1234567", "The CURIE HP1234567 has no colon `:` or underscore `_`"),
                              (None, "Curie must not be None"))
                             )
    def test_from_curie__failures(self, curie, message):
        with pytest.raises(ValueError) as cm:
            TermId.from_curie(curie)

        assert message == cm.value.args[0]

    @pytest.mark.parametrize('curie, prefix, id',
                             (
                                     ["HP:1234567", "HP", "1234567"],
                                     ["HP_1234567", "HP", "1234567"]
                             )
                             )
    def test_properties(self, curie, prefix, id):
        term_id = TermId.from_curie(curie)

        assert prefix == term_id.prefix
        assert id == term_id.id
        assert f'{prefix}:{id}' == term_id.value

    @pytest.mark.parametrize('left, right',
                             (["HP:1234567", "HP:1234567"],
                              ["HP:1234567", "HP_1234567"],
                              ["HP_1234567", "HP:1234567"],
                              ["HP_1234567", "HP_1234567"],)
                             )
    def test_equality(self, left, right):
        left = TermId.from_curie(left)
        right = TermId.from_curie(right)

        assert left == right

    @pytest.mark.parametrize('left, right, is_gt, is_eq, is_lt',
                             (
                                     # First compare by prefix
                                     ["AA:0000000", "BB:0000000", False, False, True],
                                     ["BB:0000000", "BB:0000000", False, True, False],
                                     ["CC:0000000", "BB:0000000", True, False, False],
                                     # Then by value
                                     ["HP:0000001", "HP:0000002", False, False, True],
                                     ["HP:0000002", "HP:0000002", False, True, False],
                                     ["HP:0000003", "HP:0000002", True, False, False],
                             )
                             )
    def test_comparison(self, left, right, is_gt, is_eq, is_lt):
        left = TermId.from_curie(left)
        right = TermId.from_curie(right)
        assert (left > right) == is_gt
        assert (left == right) == is_eq
        assert (left < right) == is_lt



import abc
import csv
import typing
from collections import defaultdict
from datetime import datetime

from hpotk.model import TermId, MetadataAware
from hpotk.util import open_text_io_handle_for_writing, open_text_io_handle_for_reading


class AnnotationIcContainer(typing.Mapping[TermId, float], MetadataAware, metaclass=abc.ABCMeta):
    """
    A container for storing information content of item annotations.
    """

    def to_csv(self, fh: typing.Union[str, typing.IO]):
        """
        Store the term ID to IC mapping with metadata into a CSV file.
        :param fh: where to write the
        :return:
        """
        now = datetime.now()
        self.metadata['created'] = now.strftime('%Y-%m-%d-%H:%M:%S')
        with open_text_io_handle_for_writing(fh) as handle:
            # (0) Comments
            handle.write('#Information content of the term ID calculated from HPO annotations\n')
            handle.write('#' + self.metadata_to_str() + '\n')

            # (1) Header
            fieldnames = ['term_id', 'ic']
            writer = csv.DictWriter(handle, fieldnames=fieldnames)
            writer.writeheader()

            # (2) Entries
            for term_id, ic in self.items():
                writer.writerow({'term_id': term_id, 'ic': ic})


class SimpleAnnotationIcContainer(AnnotationIcContainer):
    """
    An implementation of a :class:`AnnotationIcContainer` that is backed by a :class:`dict`.
    """

    def __init__(self, data: typing.Mapping[TermId, float],
                 metadata: typing.Optional[typing.Mapping[str, str]] = None):
        if not isinstance(data, typing.Mapping):
            raise ValueError(f'data must be an instance of Mapping but it was: {type(data)}')
        self._data = data

        self._meta = dict()
        if metadata is not None:
            if not isinstance(metadata, dict):
                raise ValueError(f'meta must be a dict but was {type(metadata)}')
            else:
                self._meta.update(metadata)

    def __getitem__(self, key: TermId) -> float:
        return self._data[key]

    def __len__(self) -> int:
        return len(self._data)

    def __iter__(self) -> typing.Iterator[TermId]:
        return iter(self._data)

    @property
    def metadata(self) -> typing.MutableMapping[str, str]:
        return self._meta


class SimilarityContainer(MetadataAware, typing.Sized):
    """
    A container for pre-calculated semantic similarity results.
    """

    def __init__(self, metadata: typing.Optional[typing.Mapping[str, str]] = None):
        self._meta = dict()
        if metadata is not None:
            if not isinstance(metadata, dict):
                raise ValueError(f'meta must be a dict but was {type(metadata)}')
            else:
                self._meta.update(metadata)
        self._data = self._prepare_datadict()

    def get_similarity(self, a: str, b: str) -> float:
        """
        Get similarity of two entries `a` and `b`.

        :param a: an item, e.g. `HP:1234567`
        :param b: another item, e.g. `HP:9876543`
        :return: a non-negative semantic similarity
        """
        o, i = (a, b) if a <= b else (b, a)
        outer = self._data.get(o, None)
        if outer:
            return outer.get(i, 0.)
        else:
            return 0.

    def set_similarity(self, a: str, b: str, sim: float):
        """
        Set semantic similarity for items `a` and `b`.
        :param a: an item, e.g. `HP:1234567`
        :param b: another item, e.g. `HP:9876543`
        :param sim: a non-negative semantic similarity
        """
        if sim < 0.:
            raise ValueError(f'Similarity must be non-negative: {sim}')
        if a <= b:
            self._data[a][b] = sim
        else:
            self._data[b][a] = sim

    def items(self):
        """
        Get a generator of semantic similarities.

        Each item is a tuple with three items:
        *  left item (`str`)
        * right item (`str`)
        * similarity (`float`)
        """
        for a, vals in self._data.items():
            for b, sim in vals.items():
                yield a, b, sim

    @property
    def metadata(self) -> typing.Mapping[str, str]:
        return self._meta

    @staticmethod
    def _prepare_datadict() -> typing.MutableMapping[str, typing.MutableMapping[str, float]]:
        def inner() -> float:
            return 0.

        def outer() -> defaultdict:
            return defaultdict(inner)

        return defaultdict(outer)

    def to_csv(self, fh: typing.Union[str, typing.IO]):
        now = datetime.now()
        self._meta['created'] = now.strftime('%Y-%m-%d-%H:%M:%S')
        with open_text_io_handle_for_writing(fh) as handle:
            # (0) Comments
            handle.write('#Information content of the most informative common ancestor for term pairs\n')
            handle.write('#' + self.metadata_to_str() + '\n')

            # (1) Header
            fieldnames = ['term_a', 'term_b', 'ic_mica']
            writer = csv.DictWriter(handle, fieldnames=fieldnames)
            writer.writeheader()

            # (2) Entries
            for left, right, sim in self.items():
                writer.writerow({'term_a': left, 'term_b': right, 'ic_mica': sim})

    @staticmethod
    def from_csv(fh: typing.Union[str, typing.IO]):
        header = []
        records = []

        def store_header(row: str) -> bool:
            if row[0] == '#':
                header.append(row)
                return False
            return True

        with open_text_io_handle_for_reading(fh) as handle:
            reader = csv.DictReader(filter(store_header, handle))
            for record in reader:
                records.append((record['term_a'], record['term_b'], float(record['ic_mica'])))

        meta = SimilarityContainer._parse_meta(header)
        data = SimilarityContainer(meta)
        for record in records:
            data.set_similarity(record[0], record[1], record[2])

        return data

    @staticmethod
    def _parse_meta(header: typing.Sequence[str]) -> typing.Mapping[str, str]:
        # Poor man's parsing.
        if len(header) < 2 or len(header[1]) < 2:
            return {}
        else:
            # The 2nd line is the metadata line, and we strip off the first and the last char (# and \n)
            return MetadataAware.metadata_from_str(header[1][1:-1])

    def __len__(self) -> int:
        return sum([len(inner) for inner in self._data.values()])



import unittest

import ddt

from ._base import EvidenceCode


@ddt.ddt
class TestEvidenceCode(unittest.TestCase):

    @ddt.unpack
    @ddt.data(
        ['IEA'],
        ['TAS'],
        ['PCS'],
    )
    def test_parse_ok(self, code):
        ec: EvidenceCode = EvidenceCode.parse(code)
        self.assertEqual(ec, EvidenceCode[code])

    def test_parse_error(self):
        ec = EvidenceCode.parse('GIBBERISH')
        self.assertIsNone(ec)




import abc
import typing

from hpotk.model import ID, CURIE_OR_TERM_ID_OR_IDENTIFIED, TERM, MINIMAL_TERM, Versioned
from hpotk.graph import GraphAware


class MinimalOntology(typing.Generic[ID, MINIMAL_TERM], GraphAware[ID], Versioned, metaclass=abc.ABCMeta):
    """
    `MinimalOntology` is a value object that mostly holds the ontology data.

    The typical way to load the ontology is by parsing Obographs JSON file:

    .. doctest:: minimal-ontology-api

      >>> import hpotk
      >>> hpo = hpotk.ontology.load.obographs.load_minimal_ontology('data/hp.toy.json')

    The ontology data comprises:

    * ontology graph as :class:`hpotk.graph.OntologyGraph`,
    * ontology concepts as :class:`hpotk.model.MinimalTerm`, and
    * the version of the data.

    The ontology acts as a Python container of term IDs, we can check if a term is in the ontology as:

    .. doctest:: minimal-ontology-api

      >>> seizure_curie = 'HP:0001250'
      >>> seizure_curie in hpo
      True

    This works for term IDs too:

    .. doctest:: minimal-ontology-api

      >>> seizure_id = hpotk.TermId.from_curie(seizure_curie)
      >>> seizure_id in hpo
      True

    The ontology has length - the number of *primary* terms:

    .. doctest:: minimal-ontology-api

      >>> len(hpo)
      393

    .. note::

      The toy HPO has only 393 terms. Real-life HPO has much more terms...

    The terms of `MinimalOntology` are instances of :class:`hpotk.model.MinimalTerm`.
    """

    @property
    @abc.abstractmethod
    def term_ids(self) -> typing.Iterator[ID]:
        """
        Get an iterator over term IDs of the primary AND obsolete ontology terms.
        """
        pass

    @property
    @abc.abstractmethod
    def terms(self) -> typing.Iterator[MINIMAL_TERM]:
        """
        Get an iterator over current terms (*not* obsolete terms).
        """
        pass

    @abc.abstractmethod
    def get_term(self, term_id: CURIE_OR_TERM_ID_OR_IDENTIFIED) -> typing.Optional[MINIMAL_TERM]:
        """
        Get the current term for a `term_id`.

        .. doctest:: minimal-ontology-api

          >>> seizure = hpo.get_term('HP:0001250')
          >>> seizure.name
          'Seizure'

        :param term_id: a CURIE `str` (e.g. 'HP:1234567'), a :class:`hpotk.model.TermId` or
         an :class:`hpotk.model.Identified` entity that represents a *current* or an *obsolete* term.
        :return: the current term or `None` if the ontology does not contain the term ID.
        """
        pass

    def get_term_name(self, term_id: CURIE_OR_TERM_ID_OR_IDENTIFIED) -> typing.Optional[str]:
        """
        Get the name of the term with a `term_id`.

        .. doctest:: minimal-ontology-api

          >>> seizure_name = hpo.get_term_name('HP:0001250')
          >>> seizure_name
          'Seizure'

        :param term_id: a CURIE `str` (e.g. 'HP:1234567'), a :class:`hpotk.model.TermId` or
         an :class:`hpotk.model.Identified` entity that represents a *current* or an *obsolete* term.
        :return: name of the term if the term is in ontology or `None` otherwise.
        """
        term = self.get_term(term_id)
        return term.name if term else None


    def __contains__(self, term_id: CURIE_OR_TERM_ID_OR_IDENTIFIED) -> bool:
        """
        Test if the ontology contains a `term_id`.

        Use :func:`get_term` if you want to use the corresponding term apart from knowing that it is there.

        .. doctest:: minimal-ontology-api

          >>> assert 'HP:0001250' in hpo
          >>> assert hpotk.TermId.from_curie('HP:0001250') in hpo
          >>> assert seizure in hpo

        :param term_id: a CURIE `str` (e.g. HP:1234567'), a :class:`hpotk.model.TermId` or
         an :class:`hpotk.model.Identified` entity that represents a *current* or an *obsolete* term.
        :return: `True` if the term is in the ontology and `False` otherwise.
        """
        return self.get_term(term_id) is not None

    @abc.abstractmethod
    def __len__(self):
        """
        Get the number of the primary (non-obsolete) terms in the ontology.
        
        .. doctest:: minimal-ontology-api

          >>> len(hpo)
          393

        :return: the number of primary terms
        """
        pass


class Ontology(MinimalOntology[ID, TERM], metaclass=abc.ABCMeta):
    """
    An ontology with all information available for terms.

    The terms `Ontology` are instances of :class:`hpotk.model.Term`.
    """
    pass



import unittest
import ddt

from hpotk.model import TermId

from ._simple import SimpleHpoDiseaseAnnotation


@ddt.ddt
class TestSimpleHpoAnnotation(unittest.TestCase):

    @ddt.unpack
    @ddt.data(
        [-1,  1, "Numerator -1 must be a non-negative `int`"],
        [1,  0, "Denominator 0 must be a positive `int`"],
        [1, -1, "Denominator -1 must be a positive `int`"],
    )
    def test_errors(self, numerator, denominator, msg):
        tid = TermId.from_curie('HP:1234567')
        with self.assertRaises(ValueError) as eh:
            SimpleHpoDiseaseAnnotation(tid, numerator, denominator, (), ())

        self.assertEqual(eh.exception.args[0], msg)



from ._ic import calculate_ic_for_annotated_items
from ._model import AnnotationIcContainer, SimpleAnnotationIcContainer, SimilarityContainer
from ._resnik import precalculate_ic_mica_for_hpo_concept_pairs

__all__ = [
    'calculate_ic_for_annotated_items',
    'AnnotationIcContainer', 'SimpleAnnotationIcContainer', 'SimilarityContainer',
    'precalculate_ic_mica_for_hpo_concept_pairs'
]



from hpotk.model import *


class TestMinimalTerm:

    def test_equal_terms_are_equal(self):
        one = MinimalTerm.create_minimal_term(TermId.from_curie("HP:1234567"),
                                              "First",
                                              [TermId.from_curie("HP:1111111")],
                                              False)
        other = MinimalTerm.create_minimal_term(TermId.from_curie("HP:1234567"),
                                                "First",
                                                [TermId.from_curie("HP:1111111")],
                                                False)
        assert one == other

    def test_unequal_terms_are_not_equal(self):
        one = MinimalTerm.create_minimal_term(TermId.from_curie("HP:1234567"),
                                              "First",
                                              [TermId.from_curie("HP:1111111")],
                                              False)
        other = MinimalTerm.create_minimal_term(TermId.from_curie("HP:1234567"),
                                                "Second",
                                                [TermId.from_curie("HP:1111111")],
                                                False)
        assert one != other


class TestTerm:

    def test_equal_terms_are_equal(self):
        one_id = TermId.from_curie("HP:1234567")
        one_syn = Synonym('F', SynonymCategory.EXACT, None, [TermId.from_curie('ORCID:1234-5678-1234-5678')])
        one = Term.create_term(identifier=one_id,
                               name="First",
                               alt_term_ids=[TermId.from_curie("HP:1111111")],
                               is_obsolete=False,
                               definition="First term definition",
                               comment="First comment",
                               synonyms=[one_syn], xrefs=[TermId.from_curie("SNOMED_CT:123456")])
        two_id = TermId.from_curie("HP:1234567")
        two = Term.create_term(identifier=two_id,
                               name="First",
                               alt_term_ids=[TermId.from_curie("HP:1111111")],
                               is_obsolete=False,
                               definition=Definition("First term definition", ()),
                               comment="First comment",
                               synonyms=[one_syn], xrefs=[TermId.from_curie("SNOMED_CT:123456")])
        assert one == two

    def test_not_equal_terms_are_not_equal(self):
        one_id = TermId.from_curie("HP:1111111")
        one = Term.create_term(identifier=one_id,
                               name="First",
                               alt_term_ids=[TermId.from_curie("HP:2222222")],
                               is_obsolete=False,
                               definition="First term definition",
                               comment="Other comment",
                               synonyms=None,
                               xrefs=None)
        two_id = TermId.from_curie("HP:1111111")
        two = Term.create_term(identifier=two_id,
                               name="First",
                               alt_term_ids=[TermId.from_curie("HP:2222222")],
                               is_obsolete=False,
                               definition="First term definition",
                               comment="First comment",
                               synonyms=None,
                               xrefs=None)
        assert one != two

    def test_current_obsolete_synonyms(self):
        current_one = Synonym('A', SynonymCategory.EXACT, SynonymType.LAYPERSON_TERM, None)
        current_two = Synonym('B', SynonymCategory.EXACT, SynonymType.LAYPERSON_TERM, None)
        obsolete_two = Synonym('C', SynonymCategory.EXACT, SynonymType.OBSOLETE_SYNONYM, None)
        term = Term.create_term(identifier=TermId.from_curie('HP:1111111'),
                                name="First",
                                alt_term_ids=[TermId.from_curie("HP:2222222")],
                                is_obsolete=False,
                                definition="First term definition",
                                comment="Other comment",
                                synonyms=[current_one, current_two, obsolete_two],
                                xrefs=None)

        current = list(map(lambda s: s.name, term.current_synonyms()))
        assert current == ['A', 'B']
        obsolete = list(map(lambda s: s.name, term.obsolete_synonyms()))
        assert obsolete == ['C']
